# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Impurity radiation heat sink for electron heat equation based on constant fraction of total power density."""
import dataclasses
from typing import Annotated, Literal
import chex
import jax
import jax.numpy as jnp
from torax._src import array_typing
from torax._src import math_utils
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.neoclassical.conductivity import base as conductivity_base
from torax._src.sources import base
from torax._src.sources import runtime_params as runtime_params_lib
from torax._src.sources import source as source_lib
from torax._src.sources import source_profiles as source_profiles_lib
from torax._src.sources.impurity_radiation_heat_sink import impurity_radiation_heat_sink
from torax._src.torax_pydantic import torax_pydantic


# pylint: disable=invalid-name
def radially_constant_fraction_of_Pin(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    source_name: str,
    unused_core_profiles: state.CoreProfiles,
    calculated_source_profiles: source_profiles_lib.SourceProfiles | None,
    unused_conductivity: conductivity_base.Conductivity | None,
) -> tuple[array_typing.FloatVectorCell, ...]:
  """Model function for radiation heat sink from impurities."""
  source_params = runtime_params.sources[source_name]
  assert isinstance(source_params, RuntimeParams)

  if calculated_source_profiles is None:
    raise ValueError(
        'calculated_source_profiles is a required argument for'
        ' `radially_constant_fraction_of_Pin`. This can occur if this source'
        ' function is used in an explicit source.'
    )

  # Based on source_models.sum_sources_T_e and source_models.calc_and_sum
  # sources_psi, but only summing over heating *input* sources
  # (Pohm + Paux + Palpha + ...) and summing over *both* ion + electron heating

  # TODO(b/383061556) Move away from using brittle source names to identify
  # sinks/sources.
  source_profiles = jnp.zeros_like(geo.rho)
  for source_name in calculated_source_profiles.T_e:
    if 'sink' not in source_name:
      source_profiles += calculated_source_profiles.T_e[source_name]
  for source_name in calculated_source_profiles.T_i:
    if 'sink' not in source_name:
      source_profiles += calculated_source_profiles.T_i[source_name]

  Q_total_in = source_profiles
  P_total_in = math_utils.volume_integration(Q_total_in, geo)

  # Calculate the heat sink as a fraction of the total power input
  return (
      -source_params.fraction_P_heating
      * P_total_in
      / geo.volume_face[-1]
      * jnp.ones_like(geo.rho),
  )


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams(runtime_params_lib.RuntimeParams):
  fraction_P_heating: array_typing.FloatScalar


class ImpurityRadiationHeatSinkConstantFractionConfig(base.SourceModelBase):
  """Configuration for the ImpurityRadiationHeatSink.

  Attributes:
    fraction_P_heating: Fraction of total power density to be absorbed by the
      impurity.
  """

  model_name: Annotated[
      Literal['P_in_scaled_flat_profile'], torax_pydantic.JAX_STATIC
  ] = 'P_in_scaled_flat_profile'
  fraction_P_heating: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(0.1)
  )
  mode: Annotated[runtime_params_lib.Mode, torax_pydantic.JAX_STATIC] = (
      runtime_params_lib.Mode.MODEL_BASED
  )

  def build_runtime_params(
      self,
      t: chex.Numeric,
  ) -> RuntimeParams:
    return RuntimeParams(
        prescribed_values=tuple(
            [v.get_value(t) for v in self.prescribed_values]
        ),
        mode=self.mode,
        is_explicit=self.is_explicit,
        fraction_P_heating=self.fraction_P_heating.get_value(t),
    )

  def build_source(
      self,
  ) -> impurity_radiation_heat_sink.ImpurityRadiationHeatSink:
    return impurity_radiation_heat_sink.ImpurityRadiationHeatSink(
        model_func=self.model_func
    )

  @property
  def model_func(self) -> source_lib.SourceProfileFunction:
    return radially_constant_fraction_of_Pin



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pydantic config for source models."""
import copy
from typing import Any

import immutabledict
import pydantic
from torax._src.sources import base
from torax._src.sources import bremsstrahlung_heat_sink as bremsstrahlung_heat_sink_lib
from torax._src.sources import cyclotron_radiation_heat_sink as cyclotron_radiation_heat_sink_lib
from torax._src.sources import electron_cyclotron_source as electron_cyclotron_source_lib
from torax._src.sources import fusion_heat_source as fusion_heat_source_lib
from torax._src.sources import gas_puff_source as gas_puff_source_lib
from torax._src.sources import generic_current_source as generic_current_source_lib
from torax._src.sources import generic_ion_el_heat_source as generic_ion_el_heat_source_lib
from torax._src.sources import generic_particle_source as generic_particle_source_lib
from torax._src.sources import ion_cyclotron_source as ion_cyclotron_source_lib
from torax._src.sources import ohmic_heat_source as ohmic_heat_source_lib
from torax._src.sources import pellet_source as pellet_source_lib
from torax._src.sources import qei_source as qei_source_lib
from torax._src.sources import runtime_params
from torax._src.sources import source_models
from torax._src.sources.impurity_radiation_heat_sink import impurity_radiation_constant_fraction
from torax._src.sources.impurity_radiation_heat_sink import impurity_radiation_mavrin_fit
from torax._src.torax_pydantic import torax_pydantic
from typing_extensions import Self


class Sources(torax_pydantic.BaseModelFrozen):
  """Config for source models.

  The `from_dict` method of constructing this class supports the config
  described in: https://torax.readthedocs.io/en/latest/configuration.html
  """

  ei_exchange: qei_source_lib.QeiSourceConfig = torax_pydantic.ValidatedDefault(
      {'mode': 'ZERO'}
  )
  # keep-sorted start
  bremsstrahlung: (
      bremsstrahlung_heat_sink_lib.BremsstrahlungHeatSinkConfig | None
  ) = pydantic.Field(
      discriminator='model_name',
      default=None,
  )
  cyclotron_radiation: (
      cyclotron_radiation_heat_sink_lib.CyclotronRadiationHeatSinkConfig | None
  ) = pydantic.Field(
      discriminator='model_name',
      default=None,
  )
  ecrh: electron_cyclotron_source_lib.ElectronCyclotronSourceConfig | None = (
      pydantic.Field(
          discriminator='model_name',
          default=None,
      )
  )
  fusion: fusion_heat_source_lib.FusionHeatSourceConfig | None = pydantic.Field(
      discriminator='model_name',
      default=None,
  )
  gas_puff: gas_puff_source_lib.GasPuffSourceConfig | None = pydantic.Field(
      discriminator='model_name',
      default=None,
  )
  generic_current: generic_current_source_lib.GenericCurrentSourceConfig = (
      torax_pydantic.ValidatedDefault({'mode': 'ZERO'})
  )
  generic_heat: (
      generic_ion_el_heat_source_lib.GenericIonElHeatSourceConfig | None
  ) = pydantic.Field(
      discriminator='model_name',
      default=None,
  )
  generic_particle: (
      generic_particle_source_lib.GenericParticleSourceConfig | None
  ) = pydantic.Field(
      discriminator='model_name',
      default=None,
  )
  icrh: ion_cyclotron_source_lib.IonCyclotronSourceConfig | None = (
      pydantic.Field(
          discriminator='model_name',
          default=None,
      )
  )
  impurity_radiation: (
      impurity_radiation_mavrin_fit.ImpurityRadiationHeatSinkMavrinFitConfig
      | impurity_radiation_constant_fraction.ImpurityRadiationHeatSinkConstantFractionConfig
      | None
  ) = pydantic.Field(
      discriminator='model_name',
      default=None,
  )
  ohmic: ohmic_heat_source_lib.OhmicHeatSourceConfig | None = pydantic.Field(
      discriminator='model_name',
      default=None,
  )
  pellet: pellet_source_lib.PelletSourceConfig | None = pydantic.Field(
      discriminator='model_name',
      default=None,
  )
  # keep-sorted end

  @pydantic.model_validator(mode='before')
  @classmethod
  def _set_default_model_functions(cls, x: dict[str, Any]) -> dict[str, Any]:
    constructor_data = copy.deepcopy(x)
    for k, v in x.items():
      # If this an already validated model, skip it.
      if isinstance(v, base.SourceModelBase) or v is None:
        continue
      match k:
        case 'bremsstrahlung':
          if 'model_name' not in v:
            constructor_data[k][
                'model_name'
            ] = bremsstrahlung_heat_sink_lib.DEFAULT_MODEL_FUNCTION_NAME
        case 'cyclotron_radiation':
          if 'model_name' not in v:
            constructor_data[k][
                'model_name'
            ] = cyclotron_radiation_heat_sink_lib.DEFAULT_MODEL_FUNCTION_NAME
        case 'ecrh':
          if 'model_name' not in v:
            constructor_data[k][
                'model_name'
            ] = electron_cyclotron_source_lib.DEFAULT_MODEL_FUNCTION_NAME
        case 'gas_puff':
          if 'model_name' not in v:
            constructor_data[k][
                'model_name'
            ] = gas_puff_source_lib.DEFAULT_MODEL_FUNCTION_NAME
        case 'generic_particle':
          if 'model_name' not in v:
            constructor_data[k][
                'model_name'
            ] = generic_particle_source_lib.DEFAULT_MODEL_FUNCTION_NAME
        case 'pellet':
          if 'model_name' not in v:
            constructor_data[k][
                'model_name'
            ] = pellet_source_lib.DEFAULT_MODEL_FUNCTION_NAME
        case 'fusion':
          if 'model_name' not in v:
            constructor_data[k][
                'model_name'
            ] = fusion_heat_source_lib.DEFAULT_MODEL_FUNCTION_NAME
        case 'generic_heat':
          if 'model_name' not in v:
            constructor_data[k][
                'model_name'
            ] = generic_ion_el_heat_source_lib.DEFAULT_MODEL_FUNCTION_NAME
        case 'impurity_radiation':
          if 'model_name' not in v:
            constructor_data[k][
                'model_name'
            ] = impurity_radiation_mavrin_fit.DEFAULT_MODEL_FUNCTION_NAME
        case 'icrh':
          if 'model_name' not in v:
            constructor_data[k][
                'model_name'
            ] = ion_cyclotron_source_lib.DEFAULT_MODEL_FUNCTION_NAME
        case 'ohmic':
          if 'model_name' not in v:
            constructor_data[k][
                'model_name'
            ] = ohmic_heat_source_lib.DEFAULT_MODEL_FUNCTION_NAME
    return constructor_data

  @pydantic.model_validator(mode='after')
  def validate_radiation_models(self) -> Self:
    """Validate that bremsstrahlung and Mavrin models are not both active at the same time.

    This prevents double counting radiation losses.

    Returns:
      Self for method chaining.

    Raises:
      ValueError: If both bremsstrahlung and Mavrin models are active.
    """
    # Check if both sources are defined
    if isinstance(
        self.bremsstrahlung,
        bremsstrahlung_heat_sink_lib.BremsstrahlungHeatSinkConfig,
    ) and isinstance(
        self.impurity_radiation,
        impurity_radiation_mavrin_fit.ImpurityRadiationHeatSinkMavrinFitConfig,
    ):

      bremsstrahlung_active = (
          self.bremsstrahlung.mode != runtime_params.Mode.ZERO
      )

      impurity_active = self.impurity_radiation.mode != runtime_params.Mode.ZERO

      # Only raise error if both are active (not in ZERO mode)
      if bremsstrahlung_active and impurity_active:
        raise ValueError("""
            Both bremsstrahlung and impurity_radiation
            with the Mavrin model should not be active at the same time to avoid
            double-counting Bremstrahlung losses. Please either set one of them
            to Mode.ZERO or remove one of them (most likely Bremstrahlung).
            """)

    return self

  def build_models(self) -> source_models.SourceModels:
    """Builds and returns a container with instantiated source model objects."""
    standard_sources = {}
    for k, v in dict(self).items():
      # ei_exchange is handled separately above.
      if k == 'ei_exchange':
        continue
      else:
        if v is not None:
          source = v.build_source()
          if k in standard_sources:
            raise ValueError(
                f'Trying to add another source with the same name: {k}.'
            )
          standard_sources[k] = source
    qei_source_model = self.ei_exchange.build_source()
    # Qei is a special source that is not in standard_sources.
    # It has its own attribute in SourceModels.
    return source_models.SourceModels(
        qei_source=qei_source_model,
        standard_sources=immutabledict.immutabledict(standard_sources),
    )

  @property
  def source_model_config(self) -> dict[str, base.SourceModelBase]:
    return {
        k: v
        for k, v in self.__dict__.items()
        if isinstance(v, base.SourceModelBase)
    }



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Input mapping functions for use of IMAS equilibrium IDSs with TORAX."""
import logging
import os
from typing import Any

import imas
from imas import ids_toplevel
import numpy as np
import scipy
from torax._src.geometry import geometry_loader


# pylint: disable=invalid-name
def geometry_from_IMAS(
    geometry_directory: str | None = None,
    Ip_from_parameters: bool = False,
    n_rho: int = 25,
    hires_factor: int = 4,
    slice_time: float | None = None,
    slice_index: int = 0,
    equilibrium_object: ids_toplevel.IDSToplevel | None = None,
    imas_uri: str | None = None,
    imas_filepath: str | None = None,
) -> dict[str, Any]:
  """Constructs a StandardGeometryIntermediates from a IMAS equilibrium IDS.

  Currently written for COCOSv17 and DDv4.

  Args:
    geometry_directory: Directory where to find the equilibrium object. If None,
      it defaults to another dir. See `load_geo_data` implementation.
    Ip_from_parameters: If True, the Ip is taken from the parameters and the
      values in the Geometry are rescaled to match the new Ip.
    n_rho: Radial grid points (num cells)
    hires_factor: Grid refinement factor for poloidal flux <--> plasma current
      calculations.
    slice_time: Time of slice to load from IMAS IDS. If given, overrides
      slice_index.
    slice_index: Index of slice to load from IMAS IDS.
    equilibrium_object: The equilibrium IDS containing the relevant data.
    imas_uri: The IMAS uri containing the equilibrium data.
    imas_filepath: The path to the IMAS netCDF file containing the equilibrium
      data.

  Returns:
    A StandardGeometry instance based on the input file. This can then be
    used to build a StandardGeometry by passing to `build_standard_geometry`.

  Raises:
    IndexError: If `slice_index` is out of bounds for the available time slices
      in the equilibrium object.
  """
  # If the equilibrium_object is the file name, load the ids from the netCDF.
  if equilibrium_object is not None:
    equilibrium = equilibrium_object
  elif imas_uri is not None:
    equilibrium = _load_geo_data(
        imas_uri,
        geometry_directory,
    )
  elif imas_filepath is not None:
    equilibrium = _load_geo_data(
        imas_filepath,
        geometry_directory,
    )
  else:
    raise ValueError(
        "equilibrium_object must be a string (file path) or an IDS"
    )
  # TODO(b/431977390): Currently only a single time slice is used, extend to
  # support multiple time slices.
  # Convert time to index
  if slice_time is not None:
    if not np.all(equilibrium.time[:-1] <= equilibrium.time[1:]):
      sorting_indices = np.argsort(equilibrium.time)
    else:
      sorting_indices = np.arange(len(equilibrium.time))
    # Find the closest time in the IDS that is <= slice_time
    slice_index = (
        np.searchsorted(
            equilibrium.time[sorting_indices], slice_time, side="right"
        )
        - 1
    )
    if not np.allclose(
        equilibrium.time[sorting_indices][slice_index], slice_time, atol=1e-9
    ):
      logging.warning(
          "Requested t=%s not in IDS; using t=%s)",
          slice_time,
          equilibrium.time[slice_index],
      )

  if slice_index >= len(equilibrium.time_slice):
    raise IndexError(
        f"slice_index={slice_index} out of range for IDS with "
        f"{len(equilibrium.time_slice)} time slices"
    )
  IMAS_data = equilibrium.time_slice[slice_index]
  R_major = np.asarray(equilibrium.vacuum_toroidal_field.r0)
  B_0 = np.asarray(np.abs(equilibrium.vacuum_toroidal_field.b0[0]))

  # Poloidal flux.
  psi = 1 * IMAS_data.profiles_1d.psi  # Sign changed ddv4

  # Toroidal flux.
  phi = -1 * IMAS_data.profiles_1d.phi

  # Midplane radii.
  R_in = IMAS_data.profiles_1d.r_inboard
  R_out = IMAS_data.profiles_1d.r_outboard
  # toroidal field flux function
  F = -1 * IMAS_data.profiles_1d.f

  # Flux surface integrals of various geometry quantities.
  # IDS Contour integrals.
  if IMAS_data.profiles_1d.dvolume_dpsi:
    dvoldpsi = 1 * IMAS_data.profiles_1d.dvolume_dpsi  # Sign changed ddv4.
  else:
    dvoldpsi = np.gradient(
        IMAS_data.profiles_1d.volume, IMAS_data.profiles_1d.psi
    )
  # dpsi_drho_tor
  if IMAS_data.profiles_1d.dpsi_drho_tor:
    dpsidrhotor = 1 * IMAS_data.profiles_1d.dpsi_drho_tor  # Sign changed ddv4.
  else:
    rho_tor = IMAS_data.profiles_1d.rho_tor
    if not rho_tor:
      if B_0 is None or not IMAS_data.profiles_1d.phi:
        raise ValueError(
            "rho_tor not provided and cannot be calculated from given"
            " equilibrium IDS"
        )
      rho_tor = np.sqrt(IMAS_data.profiles_1d.phi / (np.pi * B_0))
    dpsidrhotor = np.gradient(IMAS_data.profiles_1d.psi, rho_tor)

  flux_surf_avg_RBp = (
      IMAS_data.profiles_1d.gm7 * dpsidrhotor / (2 * np.pi)
  )  # dpsi, C0/C1
  flux_surf_avg_R2Bp2 = (
      IMAS_data.profiles_1d.gm3 * (dpsidrhotor**2) / (4 * np.pi**2)
  )  # C4/C1
  flux_surf_avg_Bp2 = (
      IMAS_data.profiles_1d.gm2 * (dpsidrhotor**2) / (4 * np.pi**2)
  )  # C3/C1
  int_dl_over_Bp = dvoldpsi  # C1
  flux_surf_avg_1_over_R2 = IMAS_data.profiles_1d.gm1  # C2/C1

  # This branching is needed since currently not all equilibrium codes output
  # <1/R>
  if IMAS_data.profiles_1d.gm9:
    flux_surf_avg_1_over_R = IMAS_data.profiles_1d.gm9
  else:
    logging.warning(
        "Flux surface averaged <1/R> profile (gm9) not found;"
        " assuming <1/R> ≈ 1/R_major (constant)"
    )
    flux_surf_avg_1_over_R = 1 / R_major

  # jtor in TORAX is defined as the flux surface average equivalent to the
  # flux-surface current density profile. i.e.
  # jtor_torax \equiv dI/dS = dI/drhon / (dS/drhon) = dI/drhon / spr
  # spr = vpr / ( 2 * np.pi * R_major)
  # -> Ip_profile = integrate(y = spr * jtor, x= rhon, initial = 0.0)
  jtor = -1 * IMAS_data.profiles_1d.j_phi
  rhon = IMAS_data.profiles_1d.rho_tor_norm
  if not rhon:
    if B_0 is None or not IMAS_data.profiles_1d.phi:
      raise ValueError(
          "rho_tor_norm not provided and cannot be calculated from given"
          " equilibrium IDS"
      )
    rho_tor = np.sqrt(IMAS_data.profiles_1d.phi / (np.pi * B_0))
    rhon = rho_tor / rho_tor[-1]
  vpr = 4 * np.pi * phi[-1] * rhon / (F * flux_surf_avg_1_over_R2)
  spr = vpr / (2 * np.pi * R_major)
  # This Ip_profile by integration results in a discrepancy between this term
  # and the total Ip from IDS.
  Ip_profile_unscaled = scipy.integrate.cumulative_trapezoid(
      y=spr * jtor, x=rhon, initial=0.0
  )

  # Because of the discrepancy between Ip_profile[-1] (computed by integration)
  # and global_quantities.ip, here we will scale Ip_profile such that the total
  # plasma current is equal.
  Ip_total = -1 * IMAS_data.global_quantities.ip
  Ip_profile = Ip_profile_unscaled * (Ip_total / Ip_profile_unscaled[-1])

  z_magnetic_axis = np.asarray(IMAS_data.global_quantities.magnetic_axis.z)

  return {
      "Ip_from_parameters": Ip_from_parameters,
      "R_major": R_major,
      "a_minor": np.asarray(IMAS_data.boundary.minor_radius),
      "B_0": B_0,
      "psi": psi,
      "Ip_profile": Ip_profile,
      "Phi": phi,
      "R_in": R_in,
      "R_out": R_out,
      "F": F,
      "int_dl_over_Bp": int_dl_over_Bp,
      "flux_surf_avg_1_over_R": flux_surf_avg_1_over_R,
      "flux_surf_avg_1_over_R2": flux_surf_avg_1_over_R2,
      "flux_surf_avg_RBp": flux_surf_avg_RBp,
      "flux_surf_avg_R2Bp2": flux_surf_avg_R2Bp2,
      "flux_surf_avg_Bp2": flux_surf_avg_Bp2,
      "flux_surf_avg_B2": IMAS_data.profiles_1d.gm5,
      "flux_surf_avg_1_over_B2": IMAS_data.profiles_1d.gm4,
      "delta_upper_face": IMAS_data.profiles_1d.triangularity_upper,
      "delta_lower_face": IMAS_data.profiles_1d.triangularity_lower,
      "elongation": IMAS_data.profiles_1d.elongation,
      "vpr": vpr,
      "n_rho": n_rho,
      "hires_factor": hires_factor,
      "z_magnetic_axis": z_magnetic_axis,
  }


def _load_geo_data(
    uri: str,
    geometry_directory: str | None = None,
) -> ids_toplevel.IDSToplevel:
  """Loads a full equilibrium IDS for a given uri or path_name and a given ids_name."""
  geometry_directory = geometry_loader.get_geometry_dir(geometry_directory)
  uri = os.path.join(geometry_directory, uri)
  with imas.DBEntry(uri=uri, mode="r") as db:
    ids = db.get(ids_name="equilibrium")
  return ids



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains the main programmatic entry point for running a TORAX simulation.

The intended use is
```
torax_config = torax.ToraxConfig.from_dict(config_dict)
sim_outputs = torax.run_simulation(torax_config)

# Update the config to run a new simulation with different parameters.
torax_config.update(updated_fields)
new_sim_outputs = torax.run_simulation(torax_config)
```
"""
from torax._src.config import build_runtime_params
from torax._src.orchestration import initial_state as initial_state_lib
from torax._src.orchestration import run_loop
from torax._src.orchestration import sim_state
from torax._src.orchestration import step_function
from torax._src.output_tools import output
from torax._src.output_tools import post_processing
from torax._src.torax_pydantic import model_config
import xarray as xr


def prepare_simulation(
    torax_config: model_config.ToraxConfig,
) -> tuple[
    build_runtime_params.RuntimeParamsProvider,
    sim_state.ToraxSimState,
    post_processing.PostProcessedOutputs,
    step_function.SimulationStepFn,
]:
  """Prepare a TORAX simulation returning the necessary inputs for the run loop.

  Args:
    torax_config: The TORAX config to use for the simulation.

  Returns:
    A tuple containing:

      - The runtime parameters slice provider.
      - The initial state.
      - The initial post processed outputs.
      - The simulation step function.
  """
  geometry_provider = torax_config.geometry.build_provider
  physics_models = torax_config.build_physics_models()

  solver = torax_config.solver.build_solver(
      physics_models=physics_models,
  )

  runtime_params_provider = (
      build_runtime_params.RuntimeParamsProvider.from_config(torax_config)
  )

  step_fn = step_function.SimulationStepFn(
      solver=solver,
      time_step_calculator=torax_config.time_step_calculator.time_step_calculator,
      geometry_provider=geometry_provider,
      runtime_params_provider=runtime_params_provider,
  )

  if torax_config.restart and torax_config.restart.do_restart:
    initial_state, post_processed_outputs = (
        initial_state_lib.get_initial_state_and_post_processed_outputs_from_file(
            t_initial=torax_config.numerics.t_initial,
            file_restart=torax_config.restart,
            runtime_params_provider=runtime_params_provider,
            geometry_provider=geometry_provider,
            step_fn=step_fn,
        )
    )
  else:
    initial_state, post_processed_outputs = (
        initial_state_lib.get_initial_state_and_post_processed_outputs(
            t=torax_config.numerics.t_initial,
            runtime_params_provider=runtime_params_provider,
            geometry_provider=geometry_provider,
            step_fn=step_fn,
        )
    )

  return (
      runtime_params_provider,
      initial_state,
      post_processed_outputs,
      step_fn,
  )


def run_simulation(
    torax_config: model_config.ToraxConfig,
    log_timestep_info: bool = False,
    progress_bar: bool = True,
) -> tuple[xr.DataTree, output.StateHistory]:
  """Runs a TORAX simulation using the config and returns the outputs.

  Args:
    torax_config: The TORAX config to use for the simulation.
    log_timestep_info: Whether to log the timestep information.
    progress_bar: Whether to show a progress bar.

  Returns:
    A tuple of the simulation outputs in the form of a DataTree and the state
    history which is intended for helpful use with debugging as it contains
    the `CoreProfiles`, `CoreTransport`, `CoreSources`, `Geometry`, and
    `PostProcessedOutputs` dataclasses for each step of the simulation.
  """
  (
      runtime_params_provider,
      initial_state,
      post_processed_outputs,
      step_fn,
  ) = prepare_simulation(torax_config)

  state_history, post_processed_outputs_history, sim_error = run_loop.run_loop(
      runtime_params_provider=runtime_params_provider,
      initial_state=initial_state,
      initial_post_processed_outputs=post_processed_outputs,
      step_fn=step_fn,
      log_timestep_info=log_timestep_info,
      progress_bar=progress_bar,
  )

  state_history = output.StateHistory(
      state_history=state_history,
      post_processed_outputs_history=post_processed_outputs_history,
      sim_error=sim_error,
      torax_config=torax_config,
  )

  return (
      state_history.simulation_output_to_xr(),
      state_history,
  )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Update routines for core_profiles.

Set of routines that relate to updating/creating core profiles from existing
core profiles.

Includes:
- get_prescribed_core_profile_values: Updates core profiles which are not being
  evolved by PDE.
- update_core_profiles_during_step: Intra-step updates of the evolving variables
  in core profiles during solver iterations.
- update_all_core_profiles_after_step: Updates all core_profiles after a step.
  Includes the evolved variables and derived variables like q_face, psidot, etc.
- compute_boundary_conditions_for_t_plus_dt: Computes boundary conditions for
  the next timestep and returns updates to State.
"""
import dataclasses
import functools

import jax
from jax import numpy as jnp
from torax._src import array_typing
from torax._src import jax_utils
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.core_profiles import convertors
from torax._src.core_profiles import getters
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry
from torax._src.neoclassical import neoclassical_models as neoclassical_models_lib
from torax._src.physics import formulas
from torax._src.physics import psi_calculations
from torax._src.sources import source_models as source_models_lib
from torax._src.sources import source_profile_builders
from torax._src.sources import source_profiles as source_profiles_lib

_trapz = jax.scipy.integrate.trapezoid

# pylint: disable=invalid-name


def _calculate_psi_value_constraint_from_v_loop(
    dt: array_typing.FloatScalar,
    theta: array_typing.FloatScalar,
    v_loop_lcfs_t: array_typing.FloatScalar,
    v_loop_lcfs_t_plus_dt: array_typing.FloatScalar,
    psi_lcfs_t: array_typing.FloatScalar,
) -> jax.Array:
  """Calculates the value constraint on the poloidal flux for the next time step from loop voltage."""
  theta_weighted_v_loop_lcfs = (
      1 - theta
  ) * v_loop_lcfs_t + theta * v_loop_lcfs_t_plus_dt
  return psi_lcfs_t + theta_weighted_v_loop_lcfs * dt


@jax_utils.jit
def get_prescribed_core_profile_values(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
) -> dict[str, array_typing.FloatVector]:
  """Updates core profiles which are not being evolved by PDE.

  Uses same functions as for profile initialization.

  Args:
    runtime_params: Runtime parameters at t=t_initial.
    geo: Torus geometry.
    core_profiles: Core profiles dataclass to be updated

  Returns:
    Updated core profiles values on the cell grid.
  """
  # If profiles are not evolved, they can still potential be time-evolving,
  # depending on the runtime params. If so, they are updated below.
  if not runtime_params.numerics.evolve_ion_heat:
    T_i = getters.get_updated_ion_temperature(
        runtime_params.profile_conditions, geo
    ).value
  else:
    T_i = core_profiles.T_i.value
  if not runtime_params.numerics.evolve_electron_heat:
    T_e_cell_variable = getters.get_updated_electron_temperature(
        runtime_params.profile_conditions, geo
    )
    T_e = T_e_cell_variable.value
  else:
    T_e_cell_variable = core_profiles.T_e
    T_e = T_e_cell_variable.value
  if not runtime_params.numerics.evolve_density:
    n_e_cell_variable = getters.get_updated_electron_density(
        runtime_params.profile_conditions, geo
    )
  else:
    n_e_cell_variable = core_profiles.n_e
  ions = getters.get_updated_ions(
      runtime_params,
      geo,
      n_e_cell_variable,
      T_e_cell_variable,
  )
  n_e = n_e_cell_variable.value
  n_i = ions.n_i.value
  n_impurity = ions.n_impurity.value
  impurity_fractions = ions.impurity_fractions

  return {
      'T_i': T_i,
      'T_e': T_e,
      'n_e': n_e,
      'n_i': n_i,
      'n_impurity': n_impurity,
      'impurity_fractions': impurity_fractions,
      'Z_i': ions.Z_i,
      'Z_i_face': ions.Z_i_face,
      'Z_impurity': ions.Z_impurity,
      'Z_impurity_face': ions.Z_impurity_face,
      'A_i': ions.A_i,
      'A_impurity': ions.A_impurity,
      'A_impurity_face': ions.A_impurity_face,
      'Z_eff': ions.Z_eff,
      'Z_eff_face': ions.Z_eff_face,
  }


@functools.partial(jax_utils.jit, static_argnames=['evolving_names'])
def update_core_profiles_during_step(
    x_new: tuple[cell_variable.CellVariable, ...],
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
    evolving_names: tuple[str, ...],
) -> state.CoreProfiles:
  """Returns the new core profiles after updating the evolving variables.

  Intended for use during iterative solves in the step function. Only updates
  the core profiles which are being evolved by the PDE and directly derivable
  quantities like q_face, s_face. core_profile calculations which require
  sources are not updated.

  Args:
    x_new: The new values of the evolving variables.
    runtime_params: The runtime params slice.
    geo: Magnetic geometry.
    core_profiles: The old set of core plasma profiles.
    evolving_names: The names of the evolving variables.
  """

  updated_core_profiles = convertors.solver_x_tuple_to_core_profiles(
      x_new, evolving_names, core_profiles
  )

  ions = getters.get_updated_ions(
      runtime_params,
      geo,
      updated_core_profiles.n_e,
      updated_core_profiles.T_e,
  )

  return dataclasses.replace(
      updated_core_profiles,
      n_i=ions.n_i,
      n_impurity=ions.n_impurity,
      impurity_fractions=ions.impurity_fractions,
      Z_i=ions.Z_i,
      Z_i_face=ions.Z_i_face,
      Z_impurity=ions.Z_impurity,
      Z_impurity_face=ions.Z_impurity_face,
      A_i=ions.A_i,
      A_impurity=ions.A_impurity,
      A_impurity_face=ions.A_impurity_face,
      Z_eff=ions.Z_eff,
      Z_eff_face=ions.Z_eff_face,
      q_face=psi_calculations.calc_q_face(geo, updated_core_profiles.psi),
      s_face=psi_calculations.calc_s_face(geo, updated_core_profiles.psi),
  )


def update_core_and_source_profiles_after_step(
    dt: array_typing.FloatScalar,
    x_new: tuple[cell_variable.CellVariable, ...],
    runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    core_profiles_t: state.CoreProfiles,
    core_profiles_t_plus_dt: state.CoreProfiles,
    explicit_source_profiles: source_profiles_lib.SourceProfiles,
    source_models: source_models_lib.SourceModels,
    neoclassical_models: neoclassical_models_lib.NeoclassicalModels,
    evolving_names: tuple[str, ...],
) -> tuple[state.CoreProfiles, source_profiles_lib.SourceProfiles]:
  """Returns a core profiles and source profiles after the solver has finished.

  Updates the evolved variables and derived variables like q_face, psidot, etc.

  Args:
    dt: The size of the last timestep.
    x_new: The new values of the evolving variables.
    runtime_params_t_plus_dt: The runtime params slice at t=t_plus_dt.
    geo: Magnetic geometry.
    core_profiles_t: The old set of core plasma profiles.
    core_profiles_t_plus_dt: The partially new set of core plasma profiles. On
      input into this function, all prescribed profiles and used boundary
      conditions are already set. But evolving values are not.
    explicit_source_profiles: The explicit source profiles.
    source_models: The source models.
    neoclassical_models: The neoclassical models.
    evolving_names: The names of the evolving variables.

  Returns:
    A tuple of the new core profiles and the source profiles.
  """

  updated_core_profiles_t_plus_dt = convertors.solver_x_tuple_to_core_profiles(
      x_new, evolving_names, core_profiles_t_plus_dt
  )

  ions = getters.get_updated_ions(
      runtime_params_t_plus_dt,
      geo,
      updated_core_profiles_t_plus_dt.n_e,
      updated_core_profiles_t_plus_dt.T_e,
  )

  v_loop_lcfs = (
      runtime_params_t_plus_dt.profile_conditions.v_loop_lcfs  # pylint: disable=g-long-ternary
      if runtime_params_t_plus_dt.profile_conditions.use_v_loop_lcfs_boundary_condition
      else _update_v_loop_lcfs_from_psi(
          core_profiles_t.psi,
          updated_core_profiles_t_plus_dt.psi,
          dt,
      )
  )

  j_total, j_total_face, Ip_profile_face = psi_calculations.calc_j_total(
      geo,
      updated_core_profiles_t_plus_dt.psi,
  )

  # A wholly new core profiles object is defined as a guard against neglecting
  # to update one of the attributes if doing dataclasses.replace
  intermediate_core_profiles = state.CoreProfiles(
      T_i=updated_core_profiles_t_plus_dt.T_i,
      T_e=updated_core_profiles_t_plus_dt.T_e,
      psi=updated_core_profiles_t_plus_dt.psi,
      n_e=updated_core_profiles_t_plus_dt.n_e,
      n_i=ions.n_i,
      n_impurity=ions.n_impurity,
      impurity_fractions=ions.impurity_fractions,
      Z_i=ions.Z_i,
      Z_i_face=ions.Z_i_face,
      Z_impurity=ions.Z_impurity,
      Z_impurity_face=ions.Z_impurity_face,
      psidot=core_profiles_t_plus_dt.psidot,
      q_face=psi_calculations.calc_q_face(
          geo, updated_core_profiles_t_plus_dt.psi
      ),
      s_face=psi_calculations.calc_s_face(
          geo, updated_core_profiles_t_plus_dt.psi
      ),
      A_i=ions.A_i,
      A_impurity=ions.A_impurity,
      A_impurity_face=ions.A_impurity_face,
      Z_eff=ions.Z_eff,
      Z_eff_face=ions.Z_eff_face,
      v_loop_lcfs=v_loop_lcfs,
      sigma=core_profiles_t_plus_dt.sigma,  # Not yet updated
      sigma_face=core_profiles_t_plus_dt.sigma_face,  # Not yet updated
      j_total=j_total,
      j_total_face=j_total_face,
      Ip_profile_face=Ip_profile_face,
  )

  conductivity = neoclassical_models.conductivity.calculate_conductivity(
      geo, intermediate_core_profiles
  )

  intermediate_core_profiles = dataclasses.replace(
      intermediate_core_profiles,
      sigma=conductivity.sigma,
      sigma_face=conductivity.sigma_face,
  )

  # build_source_profiles calculates the union with explicit + implicit
  total_source_profiles = source_profile_builders.build_source_profiles(
      runtime_params=runtime_params_t_plus_dt,
      geo=geo,
      source_models=source_models,
      neoclassical_models=neoclassical_models,
      core_profiles=intermediate_core_profiles,
      explicit=False,
      explicit_source_profiles=explicit_source_profiles,
      conductivity=conductivity,
  )

  if (
      not runtime_params_t_plus_dt.numerics.evolve_current
      and runtime_params_t_plus_dt.profile_conditions.psidot
      is not None
  ):
    # If psidot is prescribed and current does not evolve, use prescribed value
    psidot_value = (
        runtime_params_t_plus_dt.profile_conditions.psidot
    )
  else:
    # Otherwise, calculate psidot from psi sources.
    psi_sources = total_source_profiles.total_psi_sources(geo)
    psidot_value = psi_calculations.calculate_psidot_from_psi_sources(
        psi_sources=psi_sources,
        sigma=intermediate_core_profiles.sigma,
        resistivity_multiplier=runtime_params_t_plus_dt.numerics.resistivity_multiplier,
        psi=intermediate_core_profiles.psi,
        geo=geo,
    )
  psidot = dataclasses.replace(
      core_profiles_t_plus_dt.psidot,
      value=psidot_value,
      right_face_constraint=v_loop_lcfs,
      right_face_grad_constraint=None,
  )

  core_profiles_t_plus_dt = dataclasses.replace(
      intermediate_core_profiles,
      psidot=psidot,
  )
  return core_profiles_t_plus_dt, total_source_profiles


def compute_boundary_conditions_for_t_plus_dt(
    dt: array_typing.FloatScalar,
    runtime_params_t: runtime_params_slice.RuntimeParams,
    runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
    geo_t_plus_dt: geometry.Geometry,
    core_profiles_t: state.CoreProfiles,
) -> dict[str, dict[str, jax.Array | None]]:
  """Computes boundary conditions for the next timestep and returns updates to State.

  Args:
    dt: Size of the next timestep
    runtime_params_t: Runtime parameters for the current timestep. Will not be
      used if runtime_params_t.profile_conditions.v_loop_lcfs is
      None, i.e. if the dirichlet psi boundary condition based on Ip is used
    runtime_params_t_plus_dt: Runtime parameters for the next timestep
    geo_t_plus_dt: Geometry object for the next timestep
    core_profiles_t: Core profiles at the current timestep. Will not be used if
      runtime_params_t_plus_dt.profile_conditions.v_loop_lcfs is
      None, i.e. if the dirichlet psi boundary condition based on Ip is used

  Returns:
    Mapping from State attribute names to dictionaries updating attributes of
    each CellVariable in the state. This dict can in theory recursively replace
    values in a State object.
  """
  profile_conditions_t_plus_dt = (
      runtime_params_t_plus_dt.profile_conditions
  )
  # TODO(b/390143606): Separate out the boundary condition calculation from the
  # core profile calculation.
  n_e = getters.get_updated_electron_density(
      profile_conditions_t_plus_dt, geo_t_plus_dt
  )
  n_e_right_bc = n_e.right_face_constraint

  # Used for edge calculations and input arguments have correct edge BCs.
  ions_edge = getters.get_updated_ions(
      runtime_params_t_plus_dt,
      geo_t_plus_dt,
      dataclasses.replace(
          core_profiles_t.n_e,
          right_face_constraint=profile_conditions_t_plus_dt.n_e_right_bc,
      ),
      dataclasses.replace(
          core_profiles_t.T_e,
          right_face_constraint=profile_conditions_t_plus_dt.T_e_right_bc,
      ),
  )

  Z_i_edge = ions_edge.Z_i_face[-1]
  Z_impurity_edge = ions_edge.Z_impurity_face[-1]

  dilution_factor_edge = formulas.calculate_main_ion_dilution_factor(
      Z_i_edge,
      Z_impurity_edge,
      runtime_params_t_plus_dt.plasma_composition.Z_eff_face[-1],
  )

  n_i_bound_right = n_e_right_bc * dilution_factor_edge
  n_impurity_bound_right = (
      n_e_right_bc - n_i_bound_right * Z_i_edge
  ) / Z_impurity_edge

  return {
      'T_i': dict(
          left_face_grad_constraint=jnp.zeros(()),
          right_face_grad_constraint=None,
          right_face_constraint=profile_conditions_t_plus_dt.T_i_right_bc,
      ),
      'T_e': dict(
          left_face_grad_constraint=jnp.zeros(()),
          right_face_grad_constraint=None,
          right_face_constraint=profile_conditions_t_plus_dt.T_e_right_bc,
      ),
      'n_e': dict(
          left_face_grad_constraint=jnp.zeros(()),
          right_face_grad_constraint=None,
          right_face_constraint=jnp.array(n_e_right_bc),
      ),
      'n_i': dict(
          left_face_grad_constraint=jnp.zeros(()),
          right_face_grad_constraint=None,
          right_face_constraint=jnp.array(n_i_bound_right),
      ),
      'n_impurity': dict(
          left_face_grad_constraint=jnp.zeros(()),
          right_face_grad_constraint=None,
          right_face_constraint=jnp.array(n_impurity_bound_right),
      ),
      'psi': dict(
          right_face_grad_constraint=(
              psi_calculations.calculate_psi_grad_constraint_from_Ip(  # pylint: disable=g-long-ternary
                  Ip=profile_conditions_t_plus_dt.Ip,
                  geo=geo_t_plus_dt,
              )
              if not runtime_params_t.profile_conditions.use_v_loop_lcfs_boundary_condition
              else None
          ),
          right_face_constraint=(
              _calculate_psi_value_constraint_from_v_loop(  # pylint: disable=g-long-ternary
                  dt=dt,
                  v_loop_lcfs_t=runtime_params_t.profile_conditions.v_loop_lcfs,
                  v_loop_lcfs_t_plus_dt=profile_conditions_t_plus_dt.v_loop_lcfs,
                  psi_lcfs_t=core_profiles_t.psi.right_face_constraint,
                  theta=runtime_params_t.solver.theta_implicit,
              )
              if runtime_params_t.profile_conditions.use_v_loop_lcfs_boundary_condition
              else None
          ),
      ),
      'Z_i_edge': Z_i_edge,
      'Z_impurity_edge': Z_impurity_edge,
  }


def provide_core_profiles_t_plus_dt(
    dt: jax.Array,
    runtime_params_t: runtime_params_slice.RuntimeParams,
    runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
    geo_t_plus_dt: geometry.Geometry,
    core_profiles_t: state.CoreProfiles,
) -> state.CoreProfiles:
  """Provides state at t_plus_dt with new boundary conditions and prescribed profiles."""
  updated_boundary_conditions = compute_boundary_conditions_for_t_plus_dt(
      dt=dt,
      runtime_params_t=runtime_params_t,
      runtime_params_t_plus_dt=runtime_params_t_plus_dt,
      geo_t_plus_dt=geo_t_plus_dt,
      core_profiles_t=core_profiles_t,
  )
  updated_values = get_prescribed_core_profile_values(
      runtime_params=runtime_params_t_plus_dt,
      geo=geo_t_plus_dt,
      core_profiles=core_profiles_t,
  )
  T_i = dataclasses.replace(
      core_profiles_t.T_i,
      value=updated_values['T_i'],
      **updated_boundary_conditions['T_i'],
  )
  T_e = dataclasses.replace(
      core_profiles_t.T_e,
      value=updated_values['T_e'],
      **updated_boundary_conditions['T_e'],
  )
  psi = dataclasses.replace(
      core_profiles_t.psi, **updated_boundary_conditions['psi']
  )
  n_e = dataclasses.replace(
      core_profiles_t.n_e,
      value=updated_values['n_e'],
      **updated_boundary_conditions['n_e'],
  )
  n_i = dataclasses.replace(
      core_profiles_t.n_i,
      value=updated_values['n_i'],
      **updated_boundary_conditions['n_i'],
  )
  n_impurity = dataclasses.replace(
      core_profiles_t.n_impurity,
      value=updated_values['n_impurity'],
      **updated_boundary_conditions['n_impurity'],
  )

  # pylint: disable=invalid-name
  # Update Z_face with boundary condition Z, needed for cases where T_e
  # is evolving and updated_prescribed_core_profiles is a no-op.
  Z_i_face = jnp.concatenate(
      [
          updated_values['Z_i_face'][:-1],
          jnp.array([updated_boundary_conditions['Z_i_edge']]),
      ],
  )
  Z_impurity_face = jnp.concatenate(
      [
          updated_values['Z_impurity_face'][:-1],
          jnp.array([updated_boundary_conditions['Z_impurity_edge']]),
      ],
  )
  # pylint: enable=invalid-name
  core_profiles_t_plus_dt = dataclasses.replace(
      core_profiles_t,
      T_i=T_i,
      T_e=T_e,
      psi=psi,
      n_e=n_e,
      n_i=n_i,
      n_impurity=n_impurity,
      impurity_fractions=updated_values['impurity_fractions'],
      Z_i=updated_values['Z_i'],
      Z_i_face=Z_i_face,
      Z_impurity=updated_values['Z_impurity'],
      Z_impurity_face=Z_impurity_face,
      A_i=updated_values['A_i'],
      A_impurity=updated_values['A_impurity'],
      A_impurity_face=updated_values['A_impurity_face'],
      Z_eff=updated_values['Z_eff'],
      Z_eff_face=updated_values['Z_eff_face'],
  )
  return core_profiles_t_plus_dt


# TODO(b/406173731): Find robust solution for underdetermination and solve this
# for general theta_implicit values.
def _update_v_loop_lcfs_from_psi(
    psi_t: cell_variable.CellVariable,
    psi_t_plus_dt: cell_variable.CellVariable,
    dt: array_typing.FloatScalar,
) -> jax.Array:
  """Updates the v_loop_lcfs for the next timestep.

  For the Ip boundary condition case, the v_loop_lcfs formula is in principle
  calculated from:

  (psi_lcfs_t_plus_dt - psi_lcfs_t) / dt =
    v_loop_lcfs_t_plus_dt*theta_implicit - v_loop_lcfs_t*(1-theta_implicit)

  However this set of equations is underdetermined. We thus restrict this
  calculation assuming a fully implicit system, i.e. theta_implicit=1, which is
  the TORAX default. Be cautious when interpreting the results of this function
  when theta_implicit != 1 (non-standard usage).

  Args:
    psi_t: The psi CellVariable at the beginning of the timestep interval.
    psi_t_plus_dt: The updated psi CellVariable for the next timestep.
    dt: The size of the last timestep.

  Returns:
    The updated v_loop_lcfs for the next timestep.
  """
  psi_lcfs_t = psi_t.face_value()[-1]
  psi_lcfs_t_plus_dt = psi_t_plus_dt.face_value()[-1]
  v_loop_lcfs_t_plus_dt = (psi_lcfs_t_plus_dt - psi_lcfs_t) / dt
  return v_loop_lcfs_t_plus_dt



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The BohmGyroBohmModel class."""
import dataclasses

import jax
from jax import numpy as jnp
from torax._src import array_typing
from torax._src import constants as constants_module
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.pedestal_model import pedestal_model as pedestal_model_lib
from torax._src.transport_model import runtime_params as runtime_params_lib
from torax._src.transport_model import transport_model as transport_model_lib


# pylint: disable=invalid-name
@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams(runtime_params_lib.RuntimeParams):
  """Runtime params for the BgB transport model."""

  chi_e_bohm_coeff: array_typing.FloatScalar
  chi_e_gyrobohm_coeff: array_typing.FloatScalar
  chi_i_bohm_coeff: array_typing.FloatScalar
  chi_i_gyrobohm_coeff: array_typing.FloatScalar
  D_face_c1: array_typing.FloatScalar
  D_face_c2: array_typing.FloatScalar
  V_face_coeff: array_typing.FloatScalar
  chi_e_bohm_multiplier: array_typing.FloatScalar
  chi_e_gyrobohm_multiplier: array_typing.FloatScalar
  chi_i_bohm_multiplier: array_typing.FloatScalar
  chi_i_gyrobohm_multiplier: array_typing.FloatScalar


class BohmGyroBohmTransportModel(transport_model_lib.TransportModel):
  """Calculates various coefficients related to particle transport according to the Bohm + gyro-Bohm Model."""

  def __init__(
      self,
  ):
    super().__init__()
    self._frozen = True

  def _call_implementation(
      self,
      transport_runtime_params: runtime_params_lib.RuntimeParams,
      runtime_params: runtime_params_slice.RuntimeParams,
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
      pedestal_model_output: pedestal_model_lib.PedestalModelOutput,
  ) -> transport_model_lib.TurbulentTransport:
    r"""Calculates transport coefficients using the BohmGyroBohm model.

    We use the implementation from Tholerus et al, Section 3.3.
    https://doi.org/10.1088/1741-4326/ad6ea2

    A description is provided in physics_models.rst.
    https://torax.readthedocs.io/en/latest/physics_models.html

    Args:
      transport_runtime_params: Input runtime parameters for this transport
        model at the current time.
      runtime_params: Input runtime parameters at the current time.
      geo: Geometry of the torus.
      core_profiles: Core plasma profiles.
      pedestal_model_output: Output of the pedestal model.

    Returns:
      coeffs: The transport coefficients
    """
    del pedestal_model_output
    # pylint: disable=invalid-name
    # Required for pytype
    assert isinstance(transport_runtime_params, RuntimeParams)

    # Bohm term of heat transport
    chi_e_B = (
        geo.r_mid_face
        * core_profiles.q_face**2
        / (
            constants_module.CONSTANTS.qe
            * geo.B_0
            * core_profiles.n_e.face_value()
        )
        * (
            jnp.abs(core_profiles.n_e.face_grad())
            * core_profiles.T_e.face_value()
            + jnp.abs(core_profiles.T_e.face_grad())
            * core_profiles.n_e.face_value()
        )
        * constants_module.CONSTANTS.keV2J
        / geo.rho_b
    )

    # Set proportionality of chi_i to chi_e according to the assumptions of the
    # Bohm model.
    chi_i_B = 2 * chi_e_B

    # Gyrobohm term of heat transport
    chi_e_gB = (
        jnp.sqrt(runtime_params.plasma_composition.main_ion.A_avg / 2)
        * jnp.sqrt(core_profiles.T_e.face_value() * 1e3)
        / geo.B_0**2
        * jnp.abs(core_profiles.T_e.face_grad() * 1e3)
        / geo.rho_b
    )

    # Set proportionality of chi_i to chi_e according to the assumptions of the
    # GyroBohm model.
    chi_i_gB = 0.5 * chi_e_gB

    # Calibrated transport coefficients
    chi_e_bohm = (
        transport_runtime_params.chi_e_bohm_coeff
        * transport_runtime_params.chi_e_bohm_multiplier
        * chi_e_B
    )
    chi_e_gyrobohm = (
        transport_runtime_params.chi_e_gyrobohm_coeff
        * transport_runtime_params.chi_e_gyrobohm_multiplier
        * chi_e_gB
    )

    chi_i_bohm = (
        transport_runtime_params.chi_i_bohm_coeff
        * transport_runtime_params.chi_i_bohm_multiplier
        * chi_i_B
    )
    chi_i_gyrobohm = (
        transport_runtime_params.chi_i_gyrobohm_coeff
        * transport_runtime_params.chi_i_gyrobohm_multiplier
        * chi_i_gB
    )

    # Total heat transport (combined contributions)
    chi_e = chi_e_gyrobohm + chi_e_bohm
    chi_i = chi_i_gyrobohm + chi_i_bohm

    # Electron diffusivity
    weighting = (
        transport_runtime_params.D_face_c1
        + (
            transport_runtime_params.D_face_c2
            - transport_runtime_params.D_face_c1
        )
        * geo.rho_face_norm
    )

    # Diffusion: d_face_el is zero on-axis by definition.
    # We add a small epsilon to the denominator to avoid cases where
    # chi_i + chi_e = 0.
    d_face_el = jnp.concatenate([
        jnp.zeros(1),
        weighting[1:]
        * chi_e[1:]
        * chi_i[1:]
        / (chi_e[1:] + chi_i[1:] + constants_module.CONSTANTS.eps),
    ])

    # Electron convectivity set proportional to the electron diffusivity
    v_face_el = transport_runtime_params.V_face_coeff * d_face_el

    return transport_model_lib.TurbulentTransport(
        chi_face_ion=chi_i,
        chi_face_el=chi_e,
        d_face_el=d_face_el,
        v_face_el=v_face_el,
        chi_face_el_bohm=chi_e_bohm,
        chi_face_el_gyrobohm=chi_e_gyrobohm,
        chi_face_ion_bohm=chi_i_bohm,
        chi_face_ion_gyrobohm=chi_i_gyrobohm,
    )

  def __hash__(self):
    # All BohmGyroBohmModels are equivalent and can hash the same
    return hash('BohmGyroBohmModel')

  def __eq__(self, other):
    return isinstance(other, BohmGyroBohmTransportModel)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functionality for building discrete linear systems.

This file is expected to be used mostly internally by `fvm` itself.

The functionality here is for constructing a description of one discrete
time step of a PDE in terms of a linear equation. In practice, the
actual expressive power of the resulting Jax expression may still be
nonlinear because the coefficients of this linear equation are Jax
expressions, not just numeric values, so nonlinear solvers like
newton_raphson_solve_block can capture nonlinear dynamics even when
each step is expressed using a matrix multiply.
"""
from typing import TypeAlias

import jax
from jax import numpy as jnp
from torax._src.fvm import block_1d_coeffs
from torax._src.fvm import cell_variable
from torax._src.fvm import convection_terms
from torax._src.fvm import diffusion_terms

AuxiliaryOutput: TypeAlias = block_1d_coeffs.AuxiliaryOutput
Block1DCoeffs: TypeAlias = block_1d_coeffs.Block1DCoeffs


def calc_c(
    x: tuple[cell_variable.CellVariable, ...],
    coeffs: Block1DCoeffs,
    convection_dirichlet_mode: str = 'ghost',
    convection_neumann_mode: str = 'ghost',
) -> tuple[jax.Array, jax.Array]:
  """Calculate C and c such that F = C x + c.

  See docstrings for `Block1DCoeff` and `implicit_solve_block` for
  more detail.

  Args:
    x: Tuple containing CellVariables for each channel. This function uses only
      their shape and their boundary conditions, not their values.
    coeffs: Coefficients defining the differential equation.
    convection_dirichlet_mode: See docstring of the `convection_terms` function,
      `dirichlet_mode` argument.
    convection_neumann_mode: See docstring of the `convection_terms` function,
      `neumann_mode` argument.

  Returns:
    c_mat: matrix C, such that F = C x + c
    c: the vector c
  """

  d_face = coeffs.d_face
  v_face = coeffs.v_face
  source_mat_cell = coeffs.source_mat_cell
  source_cell = coeffs.source_cell

  num_cells = x[0].value.shape[0]
  num_channels = len(x)
  for _, x_i in enumerate(x):
    if x_i.value.shape != (num_cells,):
      raise ValueError(
          f'Expected each x channel to have shape ({num_cells},) '
          f'but got {x_i.value.shape}.'
      )

  zero_block = jnp.zeros((num_cells, num_cells))
  zero_row_of_blocks = [zero_block] * num_channels
  zero_vec = jnp.zeros((num_cells))
  zero_block_vec = [zero_vec] * num_channels

  # Make a matrix C and vector c that will accumulate contributions from
  # diffusion, convection, and source terms.
  # C and c are both block structured, with one block per channel.
  c_mat = [zero_row_of_blocks.copy() for _ in range(num_channels)]
  c = zero_block_vec.copy()

  # Add diffusion terms
  if d_face is not None:
    for i in range(num_channels):
      (
          diffusion_mat,
          diffusion_vec,
      ) = diffusion_terms.make_diffusion_terms(
          d_face[i],
          x[i],
      )

      c_mat[i][i] += diffusion_mat
      c[i] += diffusion_vec

  # Add convection terms
  if v_face is not None:
    for i in range(num_channels):
      # Resolve diffusion to zeros if it is not specified
      d_face_i = d_face[i] if d_face is not None else None
      d_face_i = jnp.zeros_like(v_face[i]) if d_face_i is None else d_face_i

      (
          conv_mat,
          conv_vec,
      ) = convection_terms.make_convection_terms(
          v_face[i],
          d_face_i,
          x[i],
          dirichlet_mode=convection_dirichlet_mode,
          neumann_mode=convection_neumann_mode,
      )

      c_mat[i][i] += conv_mat
      c[i] += conv_vec

  # Add implicit source terms
  if source_mat_cell is not None:
    for i in range(num_channels):
      for j in range(num_channels):
        source = source_mat_cell[i][j]
        if source is not None:
          c_mat[i][j] += jnp.diag(source)

  # Add explicit source terms
  def add(left: jax.Array, right: jax.Array | None):
    """Addition with adding None treated as no-op."""
    if right is not None:
      return left + right
    return left

  if source_cell is not None:
    c = [add(c_i, source_i) for c_i, source_i in zip(c, source_cell)]

  # Form block structure
  c_mat = jnp.block(c_mat)
  c = jnp.block(c)

  return c_mat, c



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Conversions utilities for fvm objects."""

import dataclasses

import jax
from jax import numpy as jnp
from torax._src import state
from torax._src.core_profiles import convertors
from torax._src.fvm import cell_variable


def cell_variable_tuple_to_vec(
    x_tuple: tuple[cell_variable.CellVariable, ...],
) -> jax.Array:
  """Converts a tuple of CellVariables to a flat array.

  Args:
    x_tuple: A tuple of CellVariables.

  Returns:
    A flat array of evolving state variables.
  """
  x_vec = jnp.concatenate([x.value for x in x_tuple])
  return x_vec


def vec_to_cell_variable_tuple(
    x_vec: jax.Array,
    core_profiles: state.CoreProfiles,
    evolving_names: tuple[str, ...],
) -> tuple[cell_variable.CellVariable, ...]:
  """Converts a flat array of core profile state vars to CellVariable tuple.

  Requires an input CoreProfiles to provide boundary condition information.

  Args:
    x_vec: A flat array of evolving core profile state variables. The order of
      the variables in the array must match the order of the evolving_names.
    core_profiles: CoreProfiles containing all CellVariables with appropriate
      boundary conditions.
    evolving_names: The names of the evolving cell variables.

  Returns:
    A tuple of updated CellVariables.
  """
  x_split = jnp.split(x_vec, len(evolving_names))

  # First scale the core profiles to match the scaling in x_split, then
  # update the values in the scaled core profiles with new values from x_split.
  scaled_evolving_cp_list = [
      convertors.scale_cell_variable(
          getattr(core_profiles, name),
          scaling_factor=1 / convertors.SCALING_FACTORS[name],
      )
      for name in evolving_names
  ]

  x_out = [
      dataclasses.replace(
          scaled_evolving_cp,
          value=value,
      )
      for scaled_evolving_cp, value in zip(scaled_evolving_cp_list, x_split)
  ]

  return tuple(x_out)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic which controls the stepping over time of the simulation."""

import dataclasses
import functools

import jax
from torax._src import jax_utils
from torax._src import physics_models as physics_models_lib
from torax._src import state
from torax._src import xnp
from torax._src.config import build_runtime_params
from torax._src.config import numerics as numerics_lib
from torax._src.config import runtime_params_slice
from torax._src.core_profiles import convertors
from torax._src.core_profiles import getters
from torax._src.core_profiles import updaters
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry
from torax._src.geometry import geometry_provider as geometry_provider_lib
from torax._src.mhd.sawtooth import sawtooth_solver as sawtooth_solver_lib
from torax._src.orchestration import sim_state
from torax._src.output_tools import post_processing
from torax._src.physics import formulas
from torax._src.solver import solver as solver_lib
from torax._src.sources import source_profile_builders
from torax._src.sources import source_profiles as source_profiles_lib
from torax._src.time_step_calculator import time_step_calculator as ts
from torax._src.transport_model import transport_coefficients_builder

# pylint: disable=invalid-name


def check_for_errors(
    numerics: numerics_lib.Numerics,
    output_state: sim_state.ToraxSimState,
    post_processed_outputs: post_processing.PostProcessedOutputs,
) -> state.SimError:
  """Checks for errors in the simulation state."""
  if numerics.adaptive_dt:
    if output_state.solver_numeric_outputs.solver_error_state == 1:
      # Only check for min dt if the solver did not converge. Else we may have
      # converged at a dt > min_dt just before we reach min_dt.
      if output_state.dt / numerics.dt_reduction_factor < numerics.min_dt:
        return state.SimError.REACHED_MIN_DT
  state_error = output_state.check_for_errors()
  if state_error != state.SimError.NO_ERROR:
    return state_error
  else:
    return post_processed_outputs.check_for_errors()


@jax.tree_util.register_pytree_node_class
class SimulationStepFn:
  """Advances the TORAX simulation one time step.

  Unlike the Solver class, which updates certain parts of the state, a
  SimulationStepFn takes in the ToraxSimState and outputs the updated
  ToraxSimState, which contains not only the CoreProfiles but also extra
  simulation state useful for stepping as well as extra outputs useful for
  inspection inside the main run loop in `run_simulation()`. It wraps calls to
  Solver with useful features to increase robustness for convergence, like
  dt-backtracking.
  """

  def __init__(
      self,
      solver: solver_lib.Solver,
      time_step_calculator: ts.TimeStepCalculator,
      runtime_params_provider: build_runtime_params.RuntimeParamsProvider,
      geometry_provider: geometry_provider_lib.GeometryProvider,
  ):
    """Initializes the SimulationStepFn.

    Args:
      solver: Evolves the core profiles.
      time_step_calculator: Calculates the dt for each time step.
      runtime_params_provider: Object that returns a set of runtime parameters
        which may change from time step to time step or simulation run to run.
      geometry_provider: Provides the magnetic geometry for each time step based
        on the ToraxSimState at the start of the time step. The geometry may
        change from time step to time step, so the sim needs a function to
        provide which geometry to use for a given time step. A GeometryProvider
        is any callable (class or function) which takes the ToraxSimState at the
        start of a time step and returns the Geometry for that time step. For
        most use cases, only the time will be relevant from the ToraxSimState
        (in order to support time-dependent geometries).
    """
    self._solver = solver
    if self._solver.physics_models.mhd_models.sawtooth_models is not None:
      self._sawtooth_solver = sawtooth_solver_lib.SawtoothSolver(
          physics_models=self._solver.physics_models,
      )
    else:
      self._sawtooth_solver = None
    self._time_step_calculator = time_step_calculator
    self._geometry_provider = geometry_provider
    self._runtime_params_provider = runtime_params_provider

  @property
  def runtime_params_provider(
      self,
  ) -> build_runtime_params.RuntimeParamsProvider:
    return self._runtime_params_provider

  def tree_flatten(self):
    children = (
        self._runtime_params_provider,
        self._geometry_provider,
    )
    aux_data = (
        self._solver,
        self._time_step_calculator,
    )
    return children, aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    return cls(
        solver=aux_data[0],
        time_step_calculator=aux_data[1],
        runtime_params_provider=children[0],
        geometry_provider=children[1],
    )

  @property
  def geometry_provider(self) -> geometry_provider_lib.GeometryProvider:
    return self._geometry_provider

  @property
  def solver(self) -> solver_lib.Solver:
    return self._solver

  @property
  def time_step_calculator(self) -> ts.TimeStepCalculator:
    return self._time_step_calculator

  @xnp.jit
  def __call__(
      self,
      input_state: sim_state.ToraxSimState,
      previous_post_processed_outputs: post_processing.PostProcessedOutputs,
  ) -> tuple[
      sim_state.ToraxSimState,
      post_processing.PostProcessedOutputs,
  ]:
    """Advances the simulation state one time step.

      If a sawtooth model is provided, it will be checked to see if a sawtooth
    should trigger. If it does, the sawtooth model will be applied and instead
    of a full PDE solve, the step_fn will return early with a state following
    sawtooth redistribution, at a t+dt set by the sawtooth model.

    Args:
      input_state: State at the start of the time step, including the core
        profiles which are being evolved.
      previous_post_processed_outputs: Post-processed outputs from the previous
        time step.

    Returns:
      ToraxSimState containing:
        - the core profiles at the end of the time step.
        - time and time step calculator state info.
        - core_sources and core_transport at the end of the time step.
        - solver_numeric_outputs. This contains the number of iterations
          performed in the solver and the error state. The error states are:
            0 if solver converged with fine tolerance for this step
            1 if solver did not converge for this step (was above coarse tol)
            2 if solver converged within coarse tolerance. Allowed to pass with
              a warning. Occasional error=2 has low impact on final sim state.
      PostProcessedOutputs containing:
        - post-processed outputs at the end of the time step.
        - cumulative quantities.
      SimError indicating if an error has occurred during simulation.
    """
    runtime_params_t, geo_t = (
        build_runtime_params.get_consistent_runtime_params_and_geometry(
            t=input_state.t,
            runtime_params_provider=self._runtime_params_provider,
            geometry_provider=self._geometry_provider,
        )
    )

    # This only computes sources set to explicit in the
    # SourceConfig. All implicit sources will have their profiles
    # set to 0.
    explicit_source_profiles = source_profile_builders.build_source_profiles(
        runtime_params=runtime_params_t,
        geo=geo_t,
        core_profiles=input_state.core_profiles,
        source_models=self._solver.physics_models.source_models,
        neoclassical_models=self._solver.physics_models.neoclassical_models,
        explicit=True,
    )

    def _step():
      """Take either the adaptive or fixed step, depending on the config."""
      if runtime_params_t.numerics.adaptive_dt:
        return self._adaptive_step(
            runtime_params_t,
            geo_t,
            explicit_source_profiles,
            input_state,
            previous_post_processed_outputs,
        )
      else:
        return self._fixed_step(
            runtime_params_t,
            geo_t,
            explicit_source_profiles,
            input_state,
            previous_post_processed_outputs,
        )

    # If a sawtooth model is provided, and there was no previous
    # sawtooth crash, it will be checked to see if a sawtooth
    # should trigger. If it does, the sawtooth model will be applied and instead
    # of a full PDE solve, the step_fn will return early with a state following
    # sawtooth redistribution, at a t+dt set by the sawtooth model
    # configuration.
    if self._sawtooth_solver is not None:
      output_state, post_processed_outputs = xnp.cond(
          input_state.solver_numeric_outputs.sawtooth_crash,
          lambda *args: (input_state, previous_post_processed_outputs),
          self._sawtooth_step,
          runtime_params_t,
          geo_t,
          explicit_source_profiles,
          input_state,
          previous_post_processed_outputs,
      )

      output_state, post_processed_outputs = xnp.cond(
          # If the current state is a sawtooth and the previous state was not,
          # then we triggered a sawtooth crash and exit early.
          output_state.solver_numeric_outputs.sawtooth_crash
          & ~input_state.solver_numeric_outputs.sawtooth_crash,
          lambda: (output_state, post_processed_outputs),
          _step,
      )
    else:
      # If no sawtooth model is provided, take a normal step.
      output_state, post_processed_outputs = _step()

    return output_state, post_processed_outputs

  def _sawtooth_step(
      self,
      runtime_params_t: runtime_params_slice.RuntimeParams,
      geo_t: geometry.Geometry,
      explicit_source_profiles: source_profiles_lib.SourceProfiles,
      input_state: sim_state.ToraxSimState,
      previous_post_processed_outputs: post_processing.PostProcessedOutputs,
  ) -> tuple[
      sim_state.ToraxSimState,
      post_processing.PostProcessedOutputs,
  ]:
    """Performs a simulation step if a sawtooth crash is triggered."""
    assert runtime_params_t.mhd.sawtooth is not None
    dt_crash = runtime_params_t.mhd.sawtooth.crash_step_duration

    runtime_params_t_plus_crash_dt, geo_t, geo_t_plus_crash_dt = (
        _get_geo_and_runtime_params_at_t_plus_dt_and_phibdot(
            input_state.t,
            dt_crash,
            self._runtime_params_provider,
            geo_t,
            self._geometry_provider,
        )
    )

    # If no sawtooth crash is triggered, output_state and
    # post_processed_outputs will be the same as the input state and
    # previous_post_processed_outputs.
    output_state, post_processed_outputs = _sawtooth_step(
        sawtooth_solver=self._sawtooth_solver,
        runtime_params_t=runtime_params_t,
        runtime_params_t_plus_crash_dt=runtime_params_t_plus_crash_dt,
        geo_t=geo_t,
        geo_t_plus_crash_dt=geo_t_plus_crash_dt,
        explicit_source_profiles=explicit_source_profiles,
        input_state=input_state,
        input_post_processed_outputs=previous_post_processed_outputs,
    )
    return output_state, post_processed_outputs

  def step(
      self,
      dt: jax.Array,
      runtime_params_t: runtime_params_slice.RuntimeParams,
      runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
      geo_t: geometry.Geometry,
      geo_t_plus_dt: geometry.Geometry,
      input_state: sim_state.ToraxSimState,
      explicit_source_profiles: source_profiles_lib.SourceProfiles,
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      state.SolverNumericOutputs,
  ]:
    """Performs a simulation step with given dt.

    Solver may fail to converge in which case _adaptive_step() can be used to
    try smaller time step durations.

    Args:
      dt: Time step duration.
      runtime_params_t: Runtime parameters at time t.
      runtime_params_t_plus_dt: Runtime parameters at time t + dt.
      geo_t: The geometry of the torus during this time step of the simulation.
      geo_t_plus_dt: The geometry of the torus during the next time step of the
        simulation.
      input_state: State at the start of the time step, including the core
        profiles which are being evolved.
      explicit_source_profiles: Explicit source profiles computed based on the
        core profiles at the start of the time step.

    Returns:
      tuple:
        tuple of CellVariables corresponding to the evolved state variables
        SolverNumericOutputs containing error state and other solver-specific
        outputs.
    """

    core_profiles_t = input_state.core_profiles

    # Construct the CoreProfiles object for time t+dt with evolving boundary
    # conditions and time-dependent prescribed profiles not directly solved by
    # PDE system.
    core_profiles_t_plus_dt = updaters.provide_core_profiles_t_plus_dt(
        dt=dt,
        runtime_params_t=runtime_params_t,
        runtime_params_t_plus_dt=runtime_params_t_plus_dt,
        geo_t_plus_dt=geo_t_plus_dt,
        core_profiles_t=core_profiles_t,
    )

    # Initial trial for solver. If did not converge (can happen for nonlinear
    # step with large dt) we apply the adaptive time step routine if requested.
    return self._solver(
        t=input_state.t,
        dt=dt,
        runtime_params_t=runtime_params_t,
        runtime_params_t_plus_dt=runtime_params_t_plus_dt,
        geo_t=geo_t,
        geo_t_plus_dt=geo_t_plus_dt,
        core_profiles_t=core_profiles_t,
        core_profiles_t_plus_dt=core_profiles_t_plus_dt,
        explicit_source_profiles=explicit_source_profiles,
    )

  def _adaptive_step(
      self,
      runtime_params_t: runtime_params_slice.RuntimeParams,
      geo_t: geometry.Geometry,
      explicit_source_profiles: source_profiles_lib.SourceProfiles,
      input_state: sim_state.ToraxSimState,
      previous_post_processed_outputs: post_processing.PostProcessedOutputs,
  ) -> tuple[
      sim_state.ToraxSimState,
      post_processing.PostProcessedOutputs,
  ]:
    """Performs a (possibly) adaptive simulation step."""
    evolving_names = runtime_params_t.numerics.evolving_names

    initial_dt = self.time_step_calculator.next_dt(
        input_state.t,
        runtime_params_t,
        geo_t,
        input_state.core_profiles,
        input_state.core_transport,
    )

    input_type = jax.Array
    output_type = tuple[
        tuple[cell_variable.CellVariable, ...],
        jax.Array,  # dt
        state.SolverNumericOutputs,
        runtime_params_slice.RuntimeParams,
        geometry.Geometry,
        state.CoreProfiles,
    ]

    def cond_fun(inputs: tuple[input_type, output_type]):
      next_dt, output = inputs
      solver_outputs = output[2]

      # Check for NaN in the next dt to avoid a recursive loop.
      is_nan_next_dt = xnp.isnan(next_dt)

      # If the solver did not converge we need to make a new step.
      solver_did_not_converge = solver_outputs.solver_error_state == 1

      # If t + dt  is exactly the final time we may need a smaller step than
      # min_dt to exactly reach the final time.
      if runtime_params_t.numerics.exact_t_final:
        at_exact_t_final = xnp.allclose(
            input_state.t + next_dt,
            runtime_params_t.numerics.t_final,
        )
      else:
        at_exact_t_final = xnp.array(False)

      next_dt_too_small = next_dt < runtime_params_t.numerics.min_dt

      take_another_step = xnp.cond(
          solver_did_not_converge,
          # If the solver did not converge then we check if we are at the exact
          # final time and should take a smaller step. If not we also check if
          # the next dt is too small, if so we should end the step.
          lambda: xnp.cond(
              at_exact_t_final, lambda: True, lambda: ~next_dt_too_small
          ),
          lambda: False,
      )

      return take_another_step & ~is_nan_next_dt

    def body_fun(inputs: tuple[input_type, output_type]):
      dt, output = inputs
      old_solver_outputs = output[2]

      runtime_params_t_plus_dt, geo_t_with_phibdot, geo_t_plus_dt = (
          _get_geo_and_runtime_params_at_t_plus_dt_and_phibdot(
              input_state.t,
              dt,
              self._runtime_params_provider,
              geo_t,
              self._geometry_provider,
          )
      )

      core_profiles_t_plus_dt = updaters.provide_core_profiles_t_plus_dt(
          dt=dt,
          runtime_params_t=runtime_params_t,
          runtime_params_t_plus_dt=runtime_params_t_plus_dt,
          geo_t_plus_dt=geo_t_plus_dt,
          core_profiles_t=input_state.core_profiles,
      )
      # The solver returned state is still "intermediate" since the CoreProfiles
      # need to be updated by the evolved CellVariables in x_new
      x_new, solver_numeric_outputs = self._solver(
          t=input_state.t,
          dt=dt,
          runtime_params_t=runtime_params_t,
          runtime_params_t_plus_dt=runtime_params_t_plus_dt,
          geo_t=geo_t_with_phibdot,
          geo_t_plus_dt=geo_t_plus_dt,
          core_profiles_t=input_state.core_profiles,
          core_profiles_t_plus_dt=core_profiles_t_plus_dt,
          explicit_source_profiles=explicit_source_profiles,
      )
      solver_numeric_outputs = state.SolverNumericOutputs(
          solver_error_state=solver_numeric_outputs.solver_error_state,
          outer_solver_iterations=old_solver_outputs.outer_solver_iterations
          + 1,
          inner_solver_iterations=old_solver_outputs.inner_solver_iterations
          + solver_numeric_outputs.inner_solver_iterations,
          sawtooth_crash=solver_numeric_outputs.sawtooth_crash,
      )
      next_dt = dt / runtime_params_t_plus_dt.numerics.dt_reduction_factor
      return next_dt, (
          x_new,
          dt,
          solver_numeric_outputs,
          runtime_params_t_plus_dt,
          geo_t_plus_dt,
          core_profiles_t_plus_dt,
      )

    _, result = xnp.while_loop(
        cond_fun,
        body_fun,
        (
            initial_dt,
            (
                convertors.core_profiles_to_solver_x_tuple(
                    input_state.core_profiles, evolving_names
                ),
                initial_dt,
                state.SolverNumericOutputs(
                    # The solver has not converged yet as we have not performed
                    # any steps yet.
                    solver_error_state=1,
                    outer_solver_iterations=0,
                    inner_solver_iterations=0,
                    sawtooth_crash=False,
                ),
                runtime_params_t,
                geo_t,
                input_state.core_profiles,
            ),
        ),
    )
    output_state, post_processed_outputs = _finalize_outputs(
        t=input_state.t,
        dt=result[1],
        x_new=result[0],
        solver_numeric_outputs=result[2],
        runtime_params_t_plus_dt=result[3],
        geometry_t_plus_dt=result[4],
        core_profiles_t=input_state.core_profiles,
        core_profiles_t_plus_dt=result[5],
        explicit_source_profiles=explicit_source_profiles,
        physics_models=self._solver.physics_models,
        evolving_names=evolving_names,
        input_post_processed_outputs=previous_post_processed_outputs,
    )
    return output_state, post_processed_outputs

  def _fixed_step(
      self,
      runtime_params_t: runtime_params_slice.RuntimeParams,
      geo_t: geometry.Geometry,
      explicit_source_profiles: source_profiles_lib.SourceProfiles,
      input_state: sim_state.ToraxSimState,
      previous_post_processed_outputs: post_processing.PostProcessedOutputs,
  ) -> tuple[
      sim_state.ToraxSimState,
      post_processing.PostProcessedOutputs,
  ]:
    """Performs a single simulation step."""
    dt = self.time_step_calculator.next_dt(
        input_state.t,
        runtime_params_t,
        geo_t,
        input_state.core_profiles,
        input_state.core_transport,
    )

    runtime_params_t_plus_dt, geo_t, geo_t_plus_dt = (
        _get_geo_and_runtime_params_at_t_plus_dt_and_phibdot(
            input_state.t,
            dt,
            self._runtime_params_provider,
            geo_t,
            self._geometry_provider,
        )
    )
    core_profiles_t_plus_dt = updaters.provide_core_profiles_t_plus_dt(
        dt=dt,
        runtime_params_t=runtime_params_t,
        runtime_params_t_plus_dt=runtime_params_t_plus_dt,
        geo_t_plus_dt=geo_t_plus_dt,
        core_profiles_t=input_state.core_profiles,
    )
    # The solver returned state is still "intermediate" since the CoreProfiles
    # need to be updated by the evolved CellVariables in x_new
    x_new, solver_numeric_outputs = self._solver(
        t=input_state.t,
        dt=dt,
        runtime_params_t=runtime_params_t,
        runtime_params_t_plus_dt=runtime_params_t_plus_dt,
        geo_t=geo_t,
        geo_t_plus_dt=geo_t_plus_dt,
        core_profiles_t=input_state.core_profiles,
        core_profiles_t_plus_dt=core_profiles_t_plus_dt,
        explicit_source_profiles=explicit_source_profiles,
    )
    output_state, post_processed_outputs = _finalize_outputs(
        t=input_state.t,
        dt=dt,
        x_new=x_new,
        solver_numeric_outputs=solver_numeric_outputs,
        runtime_params_t_plus_dt=runtime_params_t_plus_dt,
        geometry_t_plus_dt=geo_t_plus_dt,
        core_profiles_t=input_state.core_profiles,
        core_profiles_t_plus_dt=core_profiles_t_plus_dt,
        explicit_source_profiles=explicit_source_profiles,
        physics_models=self._solver.physics_models,
        evolving_names=runtime_params_t.numerics.evolving_names,
        input_post_processed_outputs=previous_post_processed_outputs,
    )
    return output_state, post_processed_outputs


@functools.partial(
    jax_utils.jit,
    static_argnames=[
        'evolving_names',
    ],
)
def _finalize_outputs(
    t: jax.Array,
    dt: jax.Array,
    x_new: tuple[cell_variable.CellVariable, ...],
    solver_numeric_outputs: state.SolverNumericOutputs,
    geometry_t_plus_dt: geometry.Geometry,
    runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
    core_profiles_t: state.CoreProfiles,
    core_profiles_t_plus_dt: state.CoreProfiles,
    explicit_source_profiles: source_profiles_lib.SourceProfiles,
    physics_models: physics_models_lib.PhysicsModels,
    evolving_names: tuple[str, ...],
    input_post_processed_outputs: post_processing.PostProcessedOutputs,
) -> tuple[sim_state.ToraxSimState, post_processing.PostProcessedOutputs]:
  """Returns the final state and post-processed outputs."""
  final_core_profiles, final_source_profiles = (
      updaters.update_core_and_source_profiles_after_step(
          dt=dt,
          x_new=x_new,
          runtime_params_t_plus_dt=runtime_params_t_plus_dt,
          geo=geometry_t_plus_dt,
          core_profiles_t=core_profiles_t,
          core_profiles_t_plus_dt=core_profiles_t_plus_dt,
          explicit_source_profiles=explicit_source_profiles,
          source_models=physics_models.source_models,
          neoclassical_models=physics_models.neoclassical_models,
          evolving_names=evolving_names,
      )
  )
  final_total_transport = (
      transport_coefficients_builder.calculate_total_transport_coeffs(
          physics_models.pedestal_model,
          physics_models.transport_model,
          physics_models.neoclassical_models,
          runtime_params_t_plus_dt,
          geometry_t_plus_dt,
          final_core_profiles,
      )
  )

  output_state = sim_state.ToraxSimState(
      t=t + dt,
      dt=dt,
      core_profiles=final_core_profiles,
      core_sources=final_source_profiles,
      core_transport=final_total_transport,
      geometry=geometry_t_plus_dt,
      solver_numeric_outputs=solver_numeric_outputs,
  )
  post_processed_outputs = post_processing.make_post_processed_outputs(
      sim_state=output_state,
      runtime_params=runtime_params_t_plus_dt,
      previous_post_processed_outputs=input_post_processed_outputs,
  )
  return output_state, post_processed_outputs


@functools.partial(
    jax_utils.jit,
    static_argnames=[
        'sawtooth_solver',
    ],
)
def _sawtooth_step(
    *,
    sawtooth_solver: sawtooth_solver_lib.SawtoothSolver | None,
    runtime_params_t: runtime_params_slice.RuntimeParams,
    runtime_params_t_plus_crash_dt: runtime_params_slice.RuntimeParams,
    geo_t: geometry.Geometry,
    geo_t_plus_crash_dt: geometry.Geometry,
    explicit_source_profiles: source_profiles_lib.SourceProfiles,
    input_state: sim_state.ToraxSimState,
    input_post_processed_outputs: post_processing.PostProcessedOutputs,
) -> tuple[sim_state.ToraxSimState, post_processing.PostProcessedOutputs]:
  """Checks for and handles a sawtooth crash.

  If a sawtooth model is provided and a crash is triggered, this method
  computes the post-crash state and returns it. Otherwise, returns the input
  state and post-processed outputs unchanged.

  Consecutive sawtooth crashes are not allowed since standard PDE steps
  may then not take place. Therefore if the input state has sawtooth_crash set
  to True, then no crash is triggered.

  Args:
    sawtooth_solver: Sawtooth model which carries out sawtooth step..
    runtime_params_t: Runtime params at time t.
    runtime_params_t_plus_crash_dt: Runtime params at time t + crash_dt.
    geo_t: Geometry at time t.
    geo_t_plus_crash_dt: Geometry at time t + crash_dt.
    explicit_source_profiles: Explicit source profiles at time t.
    input_state: State at the start of the time step.
    input_post_processed_outputs: Post-processed outputs from the previous step.

  Returns:
    Returns a tuple (output_state, post_processed_outputs).
  """

  # Asserts needed for linter.
  assert runtime_params_t.mhd.sawtooth is not None
  assert sawtooth_solver is not None
  dt_crash = runtime_params_t.mhd.sawtooth.crash_step_duration

  # Prepare core_profiles_t_plus_crash_dt with new boundary conditions
  # and prescribed profiles if present.
  core_profiles_t_plus_crash_dt = updaters.provide_core_profiles_t_plus_dt(
      dt=dt_crash,
      runtime_params_t=runtime_params_t,
      runtime_params_t_plus_dt=runtime_params_t_plus_crash_dt,
      geo_t_plus_dt=geo_t_plus_crash_dt,
      core_profiles_t=input_state.core_profiles,
  )

  (
      x_candidate,
      solver_numeric_outputs,
  ) = sawtooth_solver(
      t=input_state.t,
      dt=dt_crash,
      runtime_params_t=runtime_params_t,
      runtime_params_t_plus_dt=runtime_params_t_plus_crash_dt,
      geo_t=geo_t,
      geo_t_plus_dt=geo_t_plus_crash_dt,
      core_profiles_t=input_state.core_profiles,
      core_profiles_t_plus_dt=core_profiles_t_plus_crash_dt,
      explicit_source_profiles=explicit_source_profiles,
  )

  def _make_post_crash_state_and_post_processed_outputs():
    """Returns the post-crash state and post-processed outputs."""

    # We also update the temperature profiles over the sawtooth time to
    # maintain constant dW/dt over the sawtooth period. While not strictly
    # realistic this avoids non-physical dW/dt=perturbations in
    # post-processing.
    # Following the sawtooth redistribution, the PDE will take over the
    # energy evolution and the physical dW/dt corresponding to the new profile
    # distribution will be calculated.
    # This must be done here and not in the sawtooth model since the Solver
    # API does not include the post-processed outputs.
    x_evolved = _evolve_x_after_sawtooth(
        x_redistributed=x_candidate,
        runtime_params_t_plus_crash_dt=runtime_params_t_plus_crash_dt,
        core_profiles_redistributed=core_profiles_t_plus_crash_dt,
        geo_t_plus_crash_dt=geo_t_plus_crash_dt,
        previous_post_processed_outputs=input_post_processed_outputs,
        evolving_names=runtime_params_t.numerics.evolving_names,
        dt_crash=dt_crash,
    )

    return _finalize_outputs(
        t=input_state.t,
        dt=dt_crash,
        x_new=x_evolved,
        solver_numeric_outputs=solver_numeric_outputs,
        runtime_params_t_plus_dt=runtime_params_t_plus_crash_dt,
        geometry_t_plus_dt=geo_t_plus_crash_dt,
        core_profiles_t=input_state.core_profiles,
        core_profiles_t_plus_dt=core_profiles_t_plus_crash_dt,
        explicit_source_profiles=explicit_source_profiles,
        physics_models=sawtooth_solver.physics_models,
        evolving_names=runtime_params_t.numerics.evolving_names,
        input_post_processed_outputs=input_post_processed_outputs,
    )

  return jax.lax.cond(
      solver_numeric_outputs.sawtooth_crash,
      _make_post_crash_state_and_post_processed_outputs,
      lambda: (
          input_state,
          input_post_processed_outputs,
      ),
  )


def _get_geo_and_runtime_params_at_t_plus_dt_and_phibdot(
    t: jax.Array,
    dt: jax.Array,
    runtime_params_provider: build_runtime_params.RuntimeParamsProvider,
    geo_t: geometry.Geometry,
    geometry_provider: geometry_provider_lib.GeometryProvider,
) -> tuple[
    runtime_params_slice.RuntimeParams,
    geometry.Geometry,
    geometry.Geometry,
]:
  """Returns the geos including Phibdot, and runtime params at t + dt.

  Args:
    t: Time at which the simulation is currently at.
    dt: Time step duration.
    runtime_params_provider: Object that returns a set of runtime parameters
      which may change from time step to time step or simulation run to run.
    geo_t: The geometry of the torus during this time step of the simulation.
    geometry_provider: Provides the magnetic geometry for each time step based
      on the ToraxSimState at the start of the time step.

  Returns:
    Tuple containing:
      - The runtime params at time t + dt.
      - The geometry of the torus during this time step of the simulation.
      - The geometry of the torus during the next time step of the simulation.
  """
  runtime_params_t_plus_dt, geo_t_plus_dt = (
      build_runtime_params.get_consistent_runtime_params_and_geometry(
          t=t + dt,
          runtime_params_provider=runtime_params_provider,
          geometry_provider=geometry_provider,
      )
  )
  if runtime_params_t_plus_dt.numerics.calcphibdot:
    geo_t, geo_t_plus_dt = geometry.update_geometries_with_Phibdot(
        dt=dt,
        geo_t=geo_t,
        geo_t_plus_dt=geo_t_plus_dt,
    )

  return (
      runtime_params_t_plus_dt,
      geo_t,
      geo_t_plus_dt,
  )


def _evolve_x_after_sawtooth(
    x_redistributed: tuple[cell_variable.CellVariable, ...],
    runtime_params_t_plus_crash_dt: runtime_params_slice.RuntimeParams,
    core_profiles_redistributed: state.CoreProfiles,
    geo_t_plus_crash_dt: geometry.Geometry,
    previous_post_processed_outputs: post_processing.PostProcessedOutputs,
    evolving_names: tuple[str, ...],
    dt_crash: jax.Array,
) -> tuple[cell_variable.CellVariable, ...]:
  """Evolves the x_redistributed after the sawtooth redistribution."""

  updated_core_profiles = convertors.solver_x_tuple_to_core_profiles(
      x_new=x_redistributed,
      evolving_names=evolving_names,
      core_profiles=core_profiles_redistributed,
  )

  ions = getters.get_updated_ions(
      runtime_params_t_plus_crash_dt,
      geo_t_plus_crash_dt,
      updated_core_profiles.n_e,
      updated_core_profiles.T_e,
  )

  updated_core_profiles = dataclasses.replace(
      updated_core_profiles,
      n_i=ions.n_i,
      n_impurity=ions.n_impurity,
  )

  (
      pressure_thermal_el,
      pressure_thermal_ion,
      pressure_thermal_tot,
  ) = formulas.calculate_pressure(updated_core_profiles)

  _, _, W_thermal_tot = formulas.calculate_stored_thermal_energy(
      pressure_thermal_el,
      pressure_thermal_ion,
      pressure_thermal_tot,
      geo_t_plus_crash_dt,
  )

  # Update temperatures to maintain constant dW/dt over the sawtooth period.
  dW_target = previous_post_processed_outputs.dW_thermal_dt * dt_crash

  factor = 1 + dW_target / W_thermal_tot

  updated_core_profiles = dataclasses.replace(
      updated_core_profiles,
      T_e=dataclasses.replace(
          updated_core_profiles.T_e,
          value=updated_core_profiles.T_e.value * factor,
      ),
      T_i=dataclasses.replace(
          updated_core_profiles.T_i,
          value=updated_core_profiles.T_i.value * factor,
      ),
  )

  x_evolved = convertors.core_profiles_to_solver_x_tuple(
      updated_core_profiles,
      evolving_names,
  )

  return x_evolved



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes and functions for defining interpolated parameters."""

import abc
from collections.abc import Mapping
import enum
from typing import Final, Literal, TypeAlias
import chex
import jax
import jax.numpy as jnp
import numpy as np
from torax._src import array_typing
from torax._src import jax_utils
import xarray as xr

RHO_NORM: Final[str] = 'rho_norm'

_interp_fn = jax_utils.jit(jnp.interp)
_interp_fn_vmap = jax_utils.jit(jax.vmap(jnp.interp, in_axes=(None, None, 1)))


@jax_utils.jit
def _step_interpolation(
    xs: array_typing.Array, x: chex.Numeric
) -> array_typing.Array:
  """Find the indices for step interpolation."""
  # For a given x, we want to find k such that self.xs[k] <= x < self.xs[k+1]
  # and return self.ys[k]. Subtracting 1 gives index k. Setting side='left'
  # means that the step occurs whenever x > self.xs. Clipping is strictly
  # necessary for the case where searchsorted returns index 0.
  return jnp.clip(jnp.searchsorted(xs, x, side='left') - 1, 0, xs.shape[0] - 1)


@enum.unique
class InterpolationMode(enum.Enum):
  """Defines how to do the interpolation.

  InterpolatedParams have many values to interpolate between, and this enum
  defines how exactly that interpolation is computed.

  Assuming inputs [x_0, ..., x_n] and [y_0, ..., y_n], for all modes, the
  interpolated param outputs y_0 for any input less than x_0 and y_n for any
  input greater than x_n.

  Options:
    PIECEWISE_LINEAR: Does piecewise-linear interpolation between the values
      provided. See numpy.interp for a longer description of how it works. (This
      uses JAX, but the behavior is the same.)
    STEP: Step-function interpolation. For any input value x in the range [x_k,
      x_k+1), the output will be y_k.
  """

  PIECEWISE_LINEAR = 'piecewise_linear'
  STEP = 'step'


InterpolationModeLiteral: TypeAlias = Literal[
    'step', 'STEP', 'piecewise_linear', 'PIECEWISE_LINEAR'
]


_ArrayOrListOfFloats: TypeAlias = array_typing.Array | list[float]

# Config input types convertible to InterpolatedParam objects.
InterpolatedVarSingleAxisInput: TypeAlias = (
    float
    | dict[float, float]
    | bool
    | dict[float, bool]
    | tuple[_ArrayOrListOfFloats, _ArrayOrListOfFloats]
    | xr.DataArray
)
InterpolatedVarTimeRhoInput: TypeAlias = (
    # Mapping from time to rho, value interpolated in rho
    Mapping[float, InterpolatedVarSingleAxisInput]
    | float
    | xr.DataArray
    | tuple[_ArrayOrListOfFloats, _ArrayOrListOfFloats, _ArrayOrListOfFloats]
    | tuple[_ArrayOrListOfFloats, _ArrayOrListOfFloats]
)


# Type-alias for a variable (in rho_norm) to be interpolated in time.
# If a string is provided, it is assumed to be an InterpolationMode else, the
# default piecewise linear interpolation is used.
TimeInterpolatedInput: TypeAlias = (
    InterpolatedVarSingleAxisInput
    | tuple[InterpolatedVarSingleAxisInput, InterpolationModeLiteral]
)
# Type-alias for a variable to be interpolated in time and rho_norm.
TimeRhoInterpolatedInput: TypeAlias = (
    InterpolatedVarTimeRhoInput
    | tuple[
        InterpolatedVarTimeRhoInput,
        Mapping[
            Literal['time_interpolation_mode', 'rho_interpolation_mode'],
            InterpolationModeLiteral,
        ],
    ]
)


class InterpolatedParamBase(abc.ABC):
  """Base class for interpolated params.

  An InterpolatedParamBase child class should implement the interface defined
  below where, given an x-value for where to interpolate, this object
  returns a value. The x-value can be either a time or spatial coordinate
  depending on what we are interpolating over.
  """

  @abc.abstractmethod
  def get_value(self, x: chex.Numeric) -> array_typing.Array:
    """Returns a value for this parameter interpolated at the given input."""


class _PiecewiseLinearInterpolatedParam(InterpolatedParamBase):
  """Parameter using piecewise-linear interpolation to compute its value."""

  def __init__(self, xs: array_typing.Array, ys: array_typing.Array):
    """Initialises a piecewise-linear interpolated param, xs must be sorted."""

    if not np.issubdtype(xs.dtype, np.floating):
      raise ValueError(f'xs must be a float array, but got {xs.dtype}.')
    if not np.issubdtype(ys.dtype, np.floating):
      raise ValueError(f'ys must be a float array, but got {ys.dtype}.')

    self._xs = xs
    self._ys = ys

    jax_utils.assert_rank(self.xs, 1)
    if self.xs.shape[0] != self.ys.shape[0]:
      raise ValueError(
          'xs and ys must have the same number of elements in the first '
          f'dimension. Given: {self.xs.shape} and {self.ys.shape}.'
      )
    if ys.ndim not in (1, 2):
      raise ValueError(f'ys must be either 1D or 2D. Given: {self.ys.shape}.')

  @property
  def xs(self) -> array_typing.Array:
    return self._xs

  @property
  def ys(self) -> array_typing.Array:
    return self._ys

  def get_value(
      self,
      x: chex.Numeric,
  ) -> array_typing.Array:
    x_shape = getattr(x, 'shape', ())
    is_jax = isinstance(x, jax.Array)
    # This function can be used inside a JITted function, where x are
    # tracers. Thus are required to use the JAX versions of functions in this
    # case.
    interp = _interp_fn if is_jax else np.interp
    full = jnp.full if is_jax else np.full

    match self.ys.ndim:
      # This is simply interp, but with fast paths for common special cases.
      case 1:
        # When ys is size 1, no interpolation is needed: all values are just
        # ys.
        if self.ys.size == 1:
          if x_shape == ():  # pylint: disable=g-explicit-bool-comparison
            return self.ys[0]
          else:
            return full(x_shape, self.ys[0], dtype=self.ys.dtype)
        else:
          return interp(x, self.xs, self.ys)
      # The 2D case is mapped across the last dimension.
      case 2:
        # Special case: no interpolation needed.
        if len(self.ys) == 1 and x_shape == ():  # pylint: disable=g-explicit-bool-comparison
          return self.ys[0]
        else:
          return _interp_fn_vmap(x, self.xs, self.ys)
      case _:
        raise ValueError(f'ys must be either 1D or 2D. Given: {self.ys.shape}.')


class _StepInterpolatedParam(InterpolatedParamBase):
  """Parameter using step interpolation to compute its value."""

  def __init__(self, xs: array_typing.Array, ys: array_typing.Array):
    """Creates a step interpolated param, xs must be sorted."""
    self._xs = jnp.asarray(xs)
    self._ys = jnp.asarray(ys)
    jax_utils.assert_rank(self.xs, 1)
    if self.ys.ndim not in (1, 2):
      raise ValueError(f'ys must be either 1D or 2D. Given: {self.ys.shape}.')
    if self.xs.shape[0] != self.ys.shape[0]:
      raise ValueError(
          'xs and ys must have the same number of elements in the first '
          f'dimension. Given: {self.xs.shape} and {self.ys.shape}.'
      )

  @property
  def xs(self) -> array_typing.Array:
    return self._xs

  @property
  def ys(self) -> array_typing.Array:
    return self._ys

  def get_value(self, x: chex.Numeric) -> array_typing.Array:
    """Returns a single value for this range at the given coordinate."""
    indices = _step_interpolation(self.xs, x)
    return self.ys[indices]


def _is_bool(
    interp_input: InterpolatedVarSingleAxisInput,
) -> bool:
  if isinstance(interp_input, dict):
    if not interp_input:
      raise ValueError('InterpolatedVarSingleAxisInput must include values.')
    value = list(interp_input.values())[0]
    return isinstance(value, bool)
  return isinstance(interp_input, bool)


def _convert_value_to_floats(
    interp_input: InterpolatedVarSingleAxisInput,
) -> InterpolatedVarSingleAxisInput:
  if isinstance(interp_input, dict):
    return {key: float(value) for key, value in interp_input.items()}
  return float(interp_input)


def convert_input_to_xs_ys(
    interp_input: TimeInterpolatedInput,
) -> tuple[np.ndarray, np.ndarray, InterpolationMode, bool]:
  """Converts config inputs into inputs suitable for constructors.

  Args:
    interp_input: The input to convert.

  Returns:
    A tuple of (xs, ys, interpolation_mode, is_bool_param) where xs and ys are
    the arrays to be used in the constructor, interpolation_mode is the
    interpolation mode to be used, and is_bool_param is True if the input is a
    bool and False otherwise.
  """
  # This function does NOT need to be jittable.
  interpolation_mode = InterpolationMode.PIECEWISE_LINEAR
  # The param is a InterpolatedVarSingleAxisInput, so we need to convert it to
  # an InterpolatedVarSingleAxis first.
  if isinstance(interp_input, tuple):
    if len(interp_input) != 2:
      raise ValueError(
          'Single axis interpolated var tuple length must be 2. The first '
          'element are the values and the second element is the '
          'interpolation mode or both values should be arrays to be directly '
          f'interpolated. Given: {interp_input}.'
      )
    if isinstance(interp_input[1], str):
      interpolation_mode = InterpolationMode[interp_input[1].upper()]
      interp_input = interp_input[0]

  if _is_bool(interp_input):
    interp_input = _convert_value_to_floats(interp_input)
    is_bool_param = True
  else:
    is_bool_param = False

  if isinstance(interp_input, xr.DataArray):
    if not isinstance(interp_input.coords, Mapping):  # pytype: disable=attribute-error
      raise ValueError('The coords in the xr.DataArray must be a mapping.')
    if 'time' not in interp_input.coords:  # pytype: disable=attribute-error
      raise ValueError(
          'The coords in the xr.DataArray must include a "time" coordinate.'
      )
    return (
        np.asarray(interp_input.coords['time'], dtype=jax_utils.get_np_dtype()),  # pytype: disable=attribute-error
        np.asarray(interp_input.values, dtype=jax_utils.get_np_dtype()),  # pytype: disable=attribute-error
        interpolation_mode,
        is_bool_param,
    )
  if isinstance(interp_input, tuple):
    if len(interp_input) != 2:
      raise ValueError(
          'The time interpolated input tuple must be length 2. Given: '
          f'{interp_input}.'
      )
    xs, ys = interp_input
    xs = np.asarray(xs, dtype=jax_utils.get_np_dtype())
    ys = np.asarray(ys, dtype=jax_utils.get_np_dtype())
    return xs, ys, interpolation_mode, is_bool_param
  if isinstance(interp_input, dict):
    if not interp_input:
      raise ValueError('The time interpolated input dict must be non-empty.')
    return (
        np.array(list(interp_input.keys()), dtype=jax_utils.get_np_dtype()),  # pytype: disable=attribute-error
        np.array(list(interp_input.values()), dtype=jax_utils.get_np_dtype()),  # pytype: disable=attribute-error
        interpolation_mode,
        is_bool_param,
    )
  else:
    # The input is a single value.
    return (
        np.array([0.0], dtype=jax_utils.get_np_dtype()),
        np.array([interp_input], dtype=jax_utils.get_np_dtype()),
        interpolation_mode,
        is_bool_param,
    )


@jax.tree_util.register_pytree_node_class
class InterpolatedVarSingleAxis(InterpolatedParamBase):
  """Parameter that may vary based on an input coordinate.

  This class is useful for defining time-dependent runtime parameters, but can
  be used to define any parameters that vary across some range.

  This function allows the interpolation of a 1d array xs, against either a 1d
  or 2d array ys. For example, xs can be time, and ys either a 1d array of
  scalars associated to the times in xs, or a 2d array where the index 0 in ys
  associates a radial array in the index 1 with the times in xs. The
  interpolation of the 2d array is then carried out element-wise and accelerated
  with vmap. Intended use of ys being a 2d array is when the radial slices on
  index 1 have already been interpolated onto appropriate TORAX grids, such as
  cell_centers, faces, or the hires grid. NOTE: this means that the 2d array
  should have shape (n, m) where n is the number of elements in the 1d array and
  m is the number of spatial grid size of the InterpolatedVar1d instance

  See `config.runtime_params.RuntimeParams` and associated tests to see how
  this is used.
  """

  def __init__(
      self,
      value: tuple[array_typing.Array, array_typing.Array],
      interpolation_mode: InterpolationMode = (
          InterpolationMode.PIECEWISE_LINEAR
      ),
      is_bool_param: bool = False,
  ):
    """Initializes InterpolatedVarSingleAxis.

    Args:
      value: A tuple of `(xs, ys)` where `xs` is assumed to be a 1D array and
        `ys` can either be a 1D or 2D array with ys.shape[0] = len(xs).
        Additionally it is expected that `xs` is sorted and an error will be
        raised at runtime if this is not the case.
      interpolation_mode: Defines how to interpolate between values in `value`.
      is_bool_param: If True, the input value is assumed to be a bool and is
        converted to a float.

    Raises:
      RuntimeError: If the input xs is not sorted.
    """
    self._value = value
    xs, ys = value
    jax_utils.error_if(xs, jnp.any(jnp.diff(xs) < 0), 'xs must be sorted.')

    if not np.issubdtype(xs.dtype, np.floating):
      raise ValueError(f'xs must be a float array, but got {xs.dtype}.')
    if not np.issubdtype(ys.dtype, np.floating):
      raise ValueError(f'ys must be a float array, but got {ys.dtype}.')

    self._is_bool_param = is_bool_param
    self._interpolation_mode = interpolation_mode
    match interpolation_mode:
      case InterpolationMode.PIECEWISE_LINEAR:
        self._param = _PiecewiseLinearInterpolatedParam(xs=xs, ys=ys)
      case InterpolationMode.STEP:
        self._param = _StepInterpolatedParam(xs=xs, ys=ys)
      case _:
        raise ValueError('Unknown interpolation mode.')

  def tree_flatten(self):
    static_params = {
        'interpolation_mode': self.interpolation_mode,
        'is_bool_param': self.is_bool_param,
    }
    return (self._value, static_params)

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    return cls(children, **aux_data)

  @property
  def is_bool_param(self) -> bool:
    """Returns whether this param represents a bool."""
    return self._is_bool_param

  @property
  def interpolation_mode(self) -> InterpolationMode:
    """Returns the interpolation mode used by this param."""
    return self._interpolation_mode

  def get_value(
      self,
      x: chex.Numeric,
  ) -> array_typing.Array:
    """Returns a single value for this range at the given coordinate."""
    value = self._param.get_value(x)
    if self._is_bool_param:
      return jnp.bool_(value > 0.5)
    return value

  @property
  def param(self) -> InterpolatedParamBase:
    """Returns the JAX-friendly interpolated param used under the hood."""
    return self._param

  def __eq__(self, other: 'InterpolatedVarSingleAxis') -> bool:
    try:
      chex.assert_trees_all_equal(self, other)
    except AssertionError:
      return False
    return True


@jax.tree_util.register_pytree_node_class
class InterpolatedVarTimeRho(InterpolatedParamBase):
  """Interpolates on a grid (time, rho).

  This class linearly interpolates along time to provide a value at any
  (time, rho) pair. For time values that are outside the range of `values` the
  closest defined `InterpolatedVarSingleAxis` is used.

  - NOTE: We assume that rho interpolation is fixed per simulation so take this
  at init and take just time at get_value.
  """

  def __init__(
      self,
      values: Mapping[float, tuple[array_typing.Array, array_typing.Array]],
      rho_norm: array_typing.Array,
      time_interpolation_mode: InterpolationMode = (
          InterpolationMode.PIECEWISE_LINEAR
      ),
      rho_interpolation_mode: InterpolationMode = (
          InterpolationMode.PIECEWISE_LINEAR
      ),
  ):
    """Constructs an `InterpolatedVarTimeRho`.

    Args:
      values: Mapping of times to (rho_norm, values) arrays of equal length.
      rho_norm: The grid to interpolate onto.
      time_interpolation_mode: The mode in which to do time interpolation.
      rho_interpolation_mode: The mode in which to do rho interpolation.
    """
    self._rho_interpolation_mode = rho_interpolation_mode
    self._time_interpolation_mode = time_interpolation_mode

    sorted_indices = np.array(sorted(values.keys()))
    rho_norm_interpolated_values = np.stack(
        [
            InterpolatedVarSingleAxis(
                values[t], rho_interpolation_mode
            ).get_value(rho_norm)
            for t in sorted_indices
        ],
        axis=0,
    )
    self._time_interpolated_var = InterpolatedVarSingleAxis(
        value=(sorted_indices, rho_norm_interpolated_values),
        interpolation_mode=time_interpolation_mode,
    )

  def tree_flatten(self):
    children = (self._time_interpolated_var,)
    aux_data = (self._rho_interpolation_mode, self._time_interpolation_mode)
    return children, aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    # Here we construct the object without calling the constructor.
    # This is because the constructor is not jittable.
    obj = object.__new__(InterpolatedVarTimeRho)
    obj._time_interpolated_var = children[0]
    obj._rho_interpolation_mode = aux_data[0]
    obj._time_interpolation_mode = aux_data[1]
    return obj

  @property
  def time_interpolation_mode(self) -> InterpolationMode:
    """Returns the time interpolation mode used by this param."""
    return self._time_interpolation_mode

  @property
  def rho_interpolation_mode(self) -> InterpolationMode:
    """Returns the rho interpolation mode used by this param."""
    return self._rho_interpolation_mode

  def get_value(self, x: chex.Numeric) -> array_typing.Array:
    """Returns the value of this parameter interpolated at x=time."""
    return self._time_interpolated_var.get_value(x)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Math operations.

Math operations that are needed for Torax, but are not specific to plasma
physics or differential equation solvers.
"""

import enum
import functools

import jax
from jax import numpy as jnp
import jaxtyping as jt
from torax._src import array_typing
from torax._src import jax_utils
from torax._src.geometry import geometry


@enum.unique
class IntegralPreservationQuantity(enum.Enum):
  """The quantity to preserve the integral of when converting to face values."""

  # Indicate that the volume integral should be preserved.
  VOLUME = 'volume'
  # Indicate that the surface integral should be preserved.
  SURFACE = 'surface'
  # Indicate that the value integral should be preserved.
  VALUE = 'value'


@functools.partial(
    jax_utils.jit,
    static_argnames=['preserved_quantity'],
)
@array_typing.jaxtyped
def cell_to_face(
    cell_values: array_typing.FloatVectorCell,
    geo: geometry.Geometry,
    preserved_quantity: IntegralPreservationQuantity = IntegralPreservationQuantity.VALUE,
) -> array_typing.FloatVectorFace:
  """Convert cell values to face values.

  We make four assumptions:
  1) Inner face values are the average of neighbouring cells.
  2) The left most face value is linearly extrapolated from the left most cell
  values.
  3) The transformation from cell to face is integration preserving.
  4) The cell spacing is constant.

  Args:
    cell_values: Values defined on the TORAX cell grid.
    geo: A geometry object.
    preserved_quantity: The quantity to preserve the integral of when converting
      to face values.

  Returns:
    Values defined on the TORAX face grid.
  """
  if len(cell_values) < 2:
    raise ValueError(
        'Cell values must have at least two values to convert to face values.'
    )
  inner_face_values = (cell_values[:-1] + cell_values[1:]) / 2.0
  # Linearly extrapolate to get left value.
  left = cell_values[0] - (inner_face_values[0] - cell_values[0])
  face_values_without_right = jnp.concatenate([left[None], inner_face_values])
  # Preserve integral.
  match preserved_quantity:
    case IntegralPreservationQuantity.VOLUME:
      diff = jnp.sum(
          cell_values * geo.vpr
      ) * geo.drho_norm - jax.scipy.integrate.trapezoid(
          face_values_without_right * geo.vpr_face[:-1], geo.rho_face_norm[:-1]
      )
      right = (
          2 * diff / geo.drho_norm
          - face_values_without_right[-1] * geo.vpr_face[-2]
      ) / geo.vpr_face[-1]
    case IntegralPreservationQuantity.SURFACE:
      diff = jnp.sum(
          cell_values * geo.spr
      ) * geo.drho_norm - jax.scipy.integrate.trapezoid(
          face_values_without_right * geo.spr_face[:-1], geo.rho_face_norm[:-1]
      )
      right = (
          2 * diff / geo.drho_norm
          - face_values_without_right[-1] * geo.spr_face[-2]
      ) / geo.spr_face[-1]
    case IntegralPreservationQuantity.VALUE:
      diff = jnp.sum(
          cell_values
      ) * geo.drho_norm - jax.scipy.integrate.trapezoid(
          face_values_without_right, geo.rho_face_norm[:-1]
      )
      right = 2 * diff / geo.drho_norm - face_values_without_right[-1]

  face_values = jnp.concatenate([face_values_without_right, right[None]])
  return face_values


@array_typing.jaxtyped
def tridiag(
    diag: jt.Shaped[array_typing.Array, 'size'],
    above: jt.Shaped[array_typing.Array, 'size-1'],
    below: jt.Shaped[array_typing.Array, 'size-1'],
) -> jt.Shaped[array_typing.Array, 'size size']:
  """Builds a tridiagonal matrix.

  Args:
    diag: The main diagonal.
    above: The +1 diagonal.
    below: The -1 diagonal.

  Returns:
    The tridiagonal matrix.
  """

  return jnp.diag(diag) + jnp.diag(above, 1) + jnp.diag(below, -1)


def cumulative_trapezoid(
    y: jax.Array,
    x: jax.Array | None = None,
    dx: float = 1.0,
    axis: int = -1,
    initial: float | None = None,
) -> jax.Array:
  """Cumulatively integrate y = f(x) using the trapezoid rule.

  JAX equivalent of scipy.integrate.cumulative_trapezoid.

  Args:
    y: array of data to integrate.
    x: optional array of sample points corresponding to the `y` values. If not
      provided, `x` defaults to equally spaced with spacing given by `dx`.
    dx: the spacing between sample points when `x` is None (default: 1.0).
    axis: the axis along which to integrate (default: -1)
    initial: a scalar value to prepend to the result. Either None (default) or
      0.0. If `initial=0`, the result is an array with the same shape as `y`. If
      ``initial=None``, the resulting array has one fewer elements than `y`
      along the `axis` dimension.

  Returns:
    The cumulative definite integral approximated by the trapezoidal rule.
  """

  if x is None:
    dx = jnp.asarray(dx, dtype=y.dtype)
  else:
    if x.ndim == 1:
      if y.shape[axis] != len(x):
        raise ValueError(
            f'The length of x is {len(x)}, but expected {y.shape[axis]}.'
        )
    else:
      if x.shape != y.shape:
        raise ValueError(
            'If x is not 1 dimensional, it must have the same shape as y.'
        )

    if x.ndim == 1:
      dx = jnp.diff(x)
      new_shape = [1] * y.ndim
      new_shape[axis] = len(dx)
      dx = jnp.reshape(dx, new_shape)
    else:
      dx = jnp.diff(x, axis=axis)

  y_sliced = functools.partial(jax.lax.slice_in_dim, y, axis=axis)

  out = jnp.cumsum(dx * (y_sliced(1, None) + y_sliced(0, -1)), axis=axis) / 2.0

  if initial is not None:
    if initial != 0.0:
      raise ValueError(
          '`initial` must be 0 or None. Non-zero values have been deprecated'
          ' since SciPy version 1.12.0.'
      )
    initial_array = jnp.asarray(initial, dtype=out.dtype)
    initial_shape = list(out.shape)
    initial_shape[axis] = 1
    initial_array = jnp.broadcast_to(initial_array, initial_shape)
    out = jnp.concatenate((initial_array, out), axis=axis)
  return out


@jax_utils.jit
@array_typing.jaxtyped
def cell_integration(
    x: array_typing.FloatVectorCell, geo: geometry.Geometry
) -> array_typing.FloatScalar:
  r"""Integrate a value `x` over the rhon grid.

  Cell variables in TORAX are defined as the average of the face values. This
  method integrates that face value over the rhon grid implicitly using the
  trapezium rule to sum the averaged face values by the face grid spacing.

  Args:
    x: The cell averaged value to integrate.
    geo: The geometry instance.

  Returns:
    Integration over the rhon grid: :math:`\int_0^1 x_{face} d\hat{rho}`
  """
  if x.shape != geo.rho_norm.shape:
    raise ValueError(
        'For cell_integration, input "x" must have same shape as the cell grid'
        f'Got x.shape={x.shape}, expected {geo.rho_norm.shape}.'
    )
  return jnp.sum(x * geo.drho_norm)


@array_typing.jaxtyped
def area_integration(
    value: array_typing.FloatVector,
    geo: geometry.Geometry,
) -> array_typing.FloatScalar:
  """Calculates integral of value using an area metric."""
  return cell_integration(value * geo.spr, geo)


@array_typing.jaxtyped
def volume_integration(
    value: array_typing.FloatVector,
    geo: geometry.Geometry,
) -> array_typing.FloatScalar:
  """Calculates integral of value using a volume metric."""
  return cell_integration(value * geo.vpr, geo)


@array_typing.jaxtyped
def line_average(
    value: array_typing.FloatVector,
    geo: geometry.Geometry,
) -> array_typing.FloatScalar:
  """Calculates line-averaged value from input profile."""
  return cell_integration(value, geo)


@array_typing.jaxtyped
def volume_average(
    value: array_typing.FloatVector,
    geo: geometry.Geometry,
) -> array_typing.FloatScalar:
  """Calculates volume-averaged value from input profile."""
  return cell_integration(value * geo.vpr, geo) / geo.volume_face[-1]



# Copyright 2024 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Common types for using jaxtyping in TORAX."""

from typing import TypeAlias, TypeVar
import jax
import jaxtyping as jt
import numpy as np
from torax._src import jax_utils
import typeguard

T = TypeVar("T")

Array: TypeAlias = jax.Array | np.ndarray

FloatScalar: TypeAlias = jt.Float[Array | float, ""]
BoolScalar: TypeAlias = jt.Bool[Array | bool, ""]
IntScalar: TypeAlias = jt.Int[Array | int, ""]

FloatVector: TypeAlias = jt.Float[Array, "_"]
BoolVector: TypeAlias = jt.Bool[Array, "_"]
FloatVectorCell: TypeAlias = jt.Float[Array, "rhon"]
FloatVectorCellPlusBoundaries: TypeAlias = jt.Float[Array, "rhon+2"]
FloatMatrixCell: TypeAlias = jt.Float[Array, "rhon rhon"]
FloatVectorFace: TypeAlias = jt.Float[Array, "rhon+1"]


def jaxtyped(fn: T) -> T:
  """Function and dataclass decorator to perform runtime type-checking.

  This will perform jaxtyping runtime type checking if the environment variable
  `TORAX_JAXTYPING` is set to "true" (default is "false").

  Args:
    fn: The function to decorate.

  Returns:
    The decorated function.
  """
  runtime_checking = jax_utils.env_bool(name="TORAX_JAXTYPING", default=False)
  if runtime_checking:
    return jt.jaxtyped(fn, typechecker=typeguard.typechecked)
  else:
    return fn



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The CombinedTransportModel class.

A class for combining transport models.
"""
import dataclasses
from typing import Callable, Sequence

import jax
import jax.numpy as jnp
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.pedestal_model import pedestal_model as pedestal_model_lib
from torax._src.transport_model import runtime_params as runtime_params_lib
from torax._src.transport_model import transport_model as transport_model_lib

# pylint: disable=protected-access


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams(runtime_params_lib.RuntimeParams):
  transport_model_params: Sequence[runtime_params_lib.RuntimeParams]
  pedestal_transport_model_params: Sequence[runtime_params_lib.RuntimeParams]


class CombinedTransportModel(transport_model_lib.TransportModel):
  """Combines coefficients from a list of transport models."""

  def __init__(
      self,
      transport_models: Sequence[transport_model_lib.TransportModel],
      pedestal_transport_models: Sequence[transport_model_lib.TransportModel],
  ):
    super().__init__()
    self.transport_models = transport_models
    self.pedestal_transport_models = pedestal_transport_models
    self._frozen = True

  def __call__(
      self,
      runtime_params: runtime_params_slice.RuntimeParams,
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
      pedestal_model_output: pedestal_model_lib.PedestalModelOutput,
  ) -> transport_model_lib.TurbulentTransport:
    if not getattr(self, "_frozen", False):
      raise RuntimeError(
          f"Subclass implementation {type(self)} forgot to "
          "freeze at the end of __init__."
      )

    transport_runtime_params = runtime_params.transport

    # Calculate the transport coefficients - includes contribution from pedestal
    # and core transport models.
    transport_coeffs = self._call_implementation(
        transport_runtime_params,
        runtime_params,
        geo,
        core_profiles,
        pedestal_model_output,
    )

    # In contrast to the base TransportModel, we do not apply domain restriction
    # as this is handled at the component model level

    # Apply min/max clipping
    transport_coeffs = self._apply_clipping(
        transport_runtime_params,
        transport_coeffs,
    )

    # In contrast to the base TransportModel, we do not apply patches, as these
    # should be handled by instantiating constant component models instead.
    # However, the rho_inner and rho_outer arguments are currently required
    # in the case where the inner/outer region are to be excluded from
    # smoothing. Smoothing is applied to
    # rho_inner < rho_norm < min(rho_ped_top, rho_outer) unless
    # smooth_everywhere is True.
    return self._smooth_coeffs(
        transport_runtime_params,
        runtime_params,
        geo,
        transport_coeffs,
        pedestal_model_output,
    )

  def _call_implementation(
      self,
      transport_runtime_params: runtime_params_lib.RuntimeParams,
      runtime_params: runtime_params_slice.RuntimeParams,
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
      pedestal_model_output: pedestal_model_lib.PedestalModelOutput,
  ) -> transport_model_lib.TurbulentTransport:
    r"""Calculates transport coefficients using the Combined model.

    Args:
      transport_runtime_params: Input runtime parameters for this
        transport model. Can change without triggering a JAX recompilation.
      runtime_params: Runtime parameters for the simulation at the current time.
      geo: Geometry of the torus at the current time.
      core_profiles: Core plasma profiles.
      pedestal_model_output: Output of the pedestal model.

    Returns:
      coeffs: The transport coefficients
    """
    # Required for pytype
    assert isinstance(transport_runtime_params, RuntimeParams)

    def apply_and_restrict(
        component_model: transport_model_lib.TransportModel,
        component_params: runtime_params_lib.RuntimeParams,
        restriction_fn: Callable[
            [
                runtime_params_lib.RuntimeParams,
                geometry.Geometry,
                transport_model_lib.TurbulentTransport,
                pedestal_model_lib.PedestalModelOutput,
            ],
            transport_model_lib.TurbulentTransport,
        ],
    ) -> transport_model_lib.TurbulentTransport:
      # TODO(b/434175682): Consider only computing transport coefficients for
      # the active domain, rather than masking them out later. This could be
      # significantly more efficient especially for pedestal models, as these
      # are only active in a small region of the domain.
      component_transport_coeffs = component_model._call_implementation(
          component_params,
          runtime_params,
          geo,
          core_profiles,
          pedestal_model_output,
      )
      component_transport_coeffs = restriction_fn(
          component_params,
          geo,
          component_transport_coeffs,
          pedestal_model_output,
      )
      return component_transport_coeffs

    pedestal_coeffs = [
        apply_and_restrict(
            model, params, self._apply_pedestal_domain_restriction
        )
        for model, params in zip(
            self.pedestal_transport_models,
            transport_runtime_params.pedestal_transport_model_params,
        )
    ]

    core_coeffs = [
        apply_and_restrict(model, params, model._apply_domain_restriction)
        for model, params in zip(
            self.transport_models,
            transport_runtime_params.transport_model_params,
        )
    ]

    # Combine the transport coefficients from core and pedestal models.
    def _combine_maybe_none_coeffs(*leaves):
      non_none_leaves = [leaf for leaf in leaves if leaf is not None]
      return sum(non_none_leaves) if non_none_leaves else None

    combined_transport_coeffs = jax.tree.map(
        _combine_maybe_none_coeffs,
        *pedestal_coeffs,
        *core_coeffs,
        # Needed to handle the case where some coefficients are None and others
        # are not.
        is_leaf=lambda x: x is None,
    )

    return combined_transport_coeffs

  def _apply_pedestal_domain_restriction(
      self,
      unused_transport_runtime_params: runtime_params_lib.RuntimeParams,
      geo: geometry.Geometry,
      transport_coeffs: transport_model_lib.TurbulentTransport,
      pedestal_model_output: pedestal_model_lib.PedestalModelOutput,
  ) -> transport_model_lib.TurbulentTransport:
    del unused_transport_runtime_params
    active_mask = geo.rho_face_norm > pedestal_model_output.rho_norm_ped_top

    chi_face_ion = jnp.where(active_mask, transport_coeffs.chi_face_ion, 0.0)
    chi_face_el = jnp.where(active_mask, transport_coeffs.chi_face_el, 0.0)
    d_face_el = jnp.where(active_mask, transport_coeffs.d_face_el, 0.0)
    v_face_el = jnp.where(active_mask, transport_coeffs.v_face_el, 0.0)

    return dataclasses.replace(
        transport_coeffs,
        chi_face_ion=chi_face_ion,
        chi_face_el=chi_face_el,
        d_face_el=d_face_el,
        v_face_el=v_face_el,
    )

  def __hash__(self):
    return hash(
        tuple(self.transport_models) + tuple(self.pedestal_transport_models)
    )

  def __eq__(self, other):
    return (
        isinstance(other, CombinedTransportModel)
        and self.transport_models == other.transport_models
        and self.pedestal_transport_models == other.pedestal_transport_models
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataclass representing runtime parameter inputs to the transport models.

This is the dataclass runtime config exposed to the user. The actual model gets
a time-interpolated version of this config via the RuntimeParams.
"""
import dataclasses

import jax
from torax._src import array_typing


# pylint: disable=invalid-name
@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams:
  """Input params for the transport model which can be used as compiled args."""

  chi_min: float
  chi_max: float
  D_e_min: float
  D_e_max: float
  V_e_min: float
  V_e_max: float
  rho_min: array_typing.FloatScalar
  rho_max: array_typing.FloatScalar
  apply_inner_patch: array_typing.BoolScalar
  D_e_inner: array_typing.FloatScalar
  V_e_inner: array_typing.FloatScalar
  chi_i_inner: array_typing.FloatScalar
  chi_e_inner: array_typing.FloatScalar
  rho_inner: array_typing.FloatScalar
  apply_outer_patch: array_typing.BoolScalar
  D_e_outer: array_typing.FloatScalar
  V_e_outer: array_typing.FloatScalar
  chi_i_outer: array_typing.FloatScalar
  chi_e_outer: array_typing.FloatScalar
  rho_outer: array_typing.FloatScalar
  smoothing_width: float
  smooth_everywhere: bool



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes for representing a circular geometry."""
import numpy as np
from torax._src.geometry import geometry
from torax._src.torax_pydantic import torax_pydantic


# Using invalid-name because we are using the same naming convention as the
# external physics implementations
# pylint: disable=invalid-name
def build_circular_geometry(
    n_rho: int,
    elongation_LCFS: float,
    R_major: float,
    a_minor: float,
    B_0: float,
    hires_factor: int,
) -> geometry.Geometry:
  """Constructs a circular Geometry instance used for testing only.

  Args:
    n_rho: Radial grid points (num cells)
    elongation_LCFS: Elongation at last closed flux surface.
    R_major: major radius (R) in meters
    a_minor: minor radius (a) in meters
    B_0: Toroidal magnetic field on axis [T]
    hires_factor: Grid refinement factor for poloidal flux <--> plasma current
      calculations.

  Returns:
    A Geometry instance.
  """
  # circular geometry assumption of r/a_minor = rho_norm, the normalized
  # toroidal flux coordinate.
  # Define mesh (Slab Uniform 1D with Jacobian = 1)
  mesh = torax_pydantic.Grid1D(nx=n_rho,)
  # toroidal flux coordinate (rho) at boundary (last closed flux surface)
  rho_b = np.asarray(a_minor)

  # normalized and unnormalized toroidal flux coordinate (rho)
  # on face and cell grids. See fvm documentation and paper for details on
  # face and cell grids.
  rho_face_norm = mesh.face_centers
  rho_norm = mesh.cell_centers
  rho_face = rho_face_norm * rho_b
  rho = rho_norm * rho_b

  R_major = np.array(R_major)
  B_0 = np.array(B_0)

  # Define toroidal flux
  Phi = np.pi * B_0 * rho**2
  Phi_face = np.pi * B_0 * rho_face**2

  # Elongation profile.
  # Set to be a linearly increasing function from 1 to elongation_LCFS, which
  # is the elongation value at the last closed flux surface, set in config.
  elongation = 1 + rho_norm * (elongation_LCFS - 1)
  elongation_face = 1 + rho_face_norm * (elongation_LCFS - 1)

  # Volume in elongated circular geometry is given by:
  # V = 2*pi^2*R*rho^2*elongation
  # S = pi*rho^2*elongation

  volume = 2 * np.pi**2 * R_major * rho**2 * elongation
  volume_face = 2 * np.pi**2 * R_major * rho_face**2 * elongation_face
  area = np.pi * rho**2 * elongation
  area_face = np.pi * rho_face**2 * elongation_face

  # V' = dV/drnorm for volume integrations
  # \nabla V = 4*pi^2*R*rho*elongation
  #   + V * (elongation_param - 1) / elongation / rho_b
  # vpr = \nabla V * rho_b
  vpr = (
      4 * np.pi**2 * R_major * rho * elongation * rho_b
      + volume / elongation * (elongation_LCFS - 1)
  )
  vpr_face = (
      4 * np.pi**2 * R_major * rho_face * elongation_face * rho_b
      + volume_face / elongation_face * (elongation_LCFS - 1)
  )
  # pylint: disable=invalid-name
  # S' = dS/drnorm for area integrals on cell grid
  spr = 2 * np.pi * rho * elongation * rho_b + area / elongation * (
      elongation_LCFS - 1
  )
  spr_face = (
      2 * np.pi * rho_face * elongation_face * rho_b
      + area_face / elongation_face * (elongation_LCFS - 1)
  )

  delta_face = np.zeros(len(rho_face))

  # Geometry variables for general geometry form of transport equations.
  # With circular geometry approximation.

  # g0: <\nabla V>
  g0 = vpr / rho_b
  g0_face = vpr_face / rho_b

  # g1: <(\nabla V)^2>
  g1 = vpr**2 / rho_b**2
  g1_face = vpr_face**2 / rho_b**2

  # g2: <(\nabla V)^2 / R^2>
  g2 = g1 / R_major**2
  g2_face = g1_face / R_major**2

  # g3: <1/R^2> (done without a elongation correction)
  # <1/R^2> =
  # 1/2pi*int_0^2pi (1/(R_major+r*cosx)^2)dx =
  # 1/( R_major^2 * (1 - (r/R_major)^2)^3/2 )
  g3 = 1 / (R_major**2 * (1 - (rho / R_major) ** 2) ** (3.0 / 2.0))
  g3_face = 1 / (R_major**2 * (1 - (rho_face / R_major) ** 2) ** (3.0 / 2.0))

  # simplifying assumption for now, for J=R*B/(R0*B_0)
  J = np.ones(len(rho))
  J_face = np.ones(len(rho_face))
  # simplified (constant) version of the F=B*R function
  F = np.ones(len(rho)) * R_major * B_0
  F_face = np.ones(len(rho_face)) * R_major * B_0

  # Using an approximation where:
  # g2g3_over_rhon = 16 * pi**4 * G2 / (J * R) where:
  # G2 = vpr / (4 * pi**2) * <1/R^2>
  # This is done due to our ad-hoc elongation assumption, which leads to more
  # reasonable values for g2g3_over_rhon through the G2 definition.
  # In the future, a more rigorous analytical geometry will be developed and
  # the direct definition of g2g3_over_rhon will be used.

  g2g3_over_rhon = 4 * np.pi**2 * vpr * g3 / (J * R_major)
  g2g3_over_rhon_face = 4 * np.pi**2 * vpr_face * g3_face / (J_face * R_major)

  # High resolution versions for j (plasma current) and psi (poloidal flux)
  # manipulations. Needed if psi is initialized from plasma current, which is
  # the only option for ad-hoc circular geometry.
  rho_hires_norm = np.linspace(0, 1, n_rho * hires_factor)
  rho_hires = rho_hires_norm * rho_b

  R_out = R_major + rho
  R_out_face = R_major + rho_face

  R_in = R_major - rho
  R_in_face = R_major - rho_face

  # assumed elongation profile on hires grid
  elongation_hires = 1 + rho_hires_norm * (elongation_LCFS - 1)

  volume_hires = 2 * np.pi**2 * R_major * rho_hires**2 * elongation_hires
  area_hires = np.pi * rho_hires**2 * elongation_hires

  # V' = dV/drnorm for volume integrations on hires grid
  vpr_hires = (
      4 * np.pi**2 * R_major * rho_hires * elongation_hires * rho_b
      + volume_hires / elongation_hires * (elongation_LCFS - 1)
  )
  # S' = dS/drnorm for area integrals on hires grid
  spr_hires = (
      2 * np.pi * rho_hires * elongation_hires * rho_b
      + area_hires / elongation_hires * (elongation_LCFS - 1)
  )

  # Analytical expressions for  <1/B^2> (gm4) and <B^2> (gm5)
  epsilon = (R_out - R_in) / (R_out + R_in)
  epsilon_face = (R_out_face - R_in_face) / (R_out_face + R_in_face)
  gm4 = B_0**-2 * (1.0 + 1.5 * epsilon**2)
  gm4_face = B_0**-2 * (1.0 + 1.5 * epsilon_face**2)
  gm5 = B_0**2 / np.sqrt(1.0 - epsilon**2)
  gm5_face = B_0**2 / np.sqrt(1.0 - epsilon_face**2)

  g3_hires = 1 / (R_major**2 * (1 - (rho_hires / R_major) ** 2) ** (3.0 / 2.0))
  F_hires = np.ones(len(rho_hires)) * B_0 * R_major
  g2g3_over_rhon_hires = 4 * np.pi**2 * vpr_hires * g3_hires * B_0 / F_hires

  return geometry.Geometry(
      # Set the standard geometry params.
      geometry_type=geometry.GeometryType.CIRCULAR,
      torax_mesh=mesh,
      Phi=Phi,
      Phi_face=Phi_face,
      R_major=R_major,
      a_minor=rho_b,
      B_0=B_0,
      volume=volume,
      volume_face=volume_face,
      area=area,
      area_face=area_face,
      vpr=vpr,
      vpr_face=vpr_face,
      spr=spr,
      spr_face=spr_face,
      delta_face=delta_face,
      g0=g0,
      g0_face=g0_face,
      g1=g1,
      g1_face=g1_face,
      g2=g2,
      g2_face=g2_face,
      g3=g3,
      g3_face=g3_face,
      gm4=gm4,
      gm4_face=gm4_face,
      gm5=gm5,
      gm5_face=gm5_face,
      g2g3_over_rhon=g2g3_over_rhon,
      g2g3_over_rhon_face=g2g3_over_rhon_face,
      g2g3_over_rhon_hires=g2g3_over_rhon_hires,
      F=F,
      F_face=F_face,
      F_hires=F_hires,
      R_in=R_in,
      R_in_face=R_in_face,
      R_out=R_out,
      R_out_face=R_out_face,
      # Set the circular geometry-specific params.
      elongation=elongation,
      elongation_face=elongation_face,
      spr_hires=spr_hires,
      rho_hires_norm=rho_hires_norm,
      rho_hires=rho_hires,
      # always initialize Phibdot as zero. It will be replaced once both geo_t
      # and geo_t_plus_dt are provided, and set to be the same for geo_t and
      # geo_t_plus_dt for each given time interval.
      Phi_b_dot=np.asarray(0.0),
      _z_magnetic_axis=np.asarray(0.0),
  )
